from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes
from RSTAB.enums import MemberShearPanelDefinitionType, MemberShearPanelFasteningArrangement

class MemberShearPanel():

    def __init__(self) -> None:
        pass

    @staticmethod
    def TrapezodialSheeting(no: int = 1,
                            name: str = '',
                            girder_length_definition: list = [True],
                            sheeting_name: str = 'FI (+) 35/207 - 0.63 (b: 1) | DIN 18807 | Fischer Profil',
                            fastening_arrangement = MemberShearPanelFasteningArrangement.FASTENING_ARRANGEMENT_EVERY_RIB,
                            panel_length: float = 2.0,
                            beam_spacing: float = 1.0,
                            coefficient_k1: float = None,
                            coefficient_k2: float = None,
                            comment: str = '',
                            params: dict = None,
                            model = Model):

        """
        Args:
            no (int): Member Shear Panel Tag
            name (str): User Defined Member Shear Panel Name
                if name == '':
                    user_defined_name_enabled = False (Automatic Name Assignment)
                else:
                    user_defined_name_enabled = True
                    name = User Defined Name
            girder_length_definition (list): Girder Length Definition List
                girder_length_definition[0] (bool): Activate/Deactivate Automatically Girder Length
                girder_length_definition[1] (float): Girder Length Value
            sheeting_name (str): Sheeting Name According to Library
            fastening_arrangement (enum): Fastening Arrangement Enumeration
            panel_length (float): Panel Length
            beam_spacing (float): Beam Spacing
            coefficient_k1 (float, optional): Coefficient K1
            coefficient_k2 (float, optional): Coefficient K2
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client Model | Member Shear Panel
        clientObject = model.clientModel.factory.create('ns0:member_shear_panel')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Member Shear Panel No.
        clientObject.no = no

        # Member Shear Panel Definition Type
        clientObject.definition_type = MemberShearPanelDefinitionType.DEFINITION_TYPE_TRAPEZOIDAL_SHEETING.name

        # Member Shear Panel User Defined Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Member Shear Panel Girder Length Definition
        if girder_length_definition[0]:
            clientObject.define_girder_length_automatically = True
        else:
            clientObject.define_girder_length_automatically = False
            clientObject.girder_length = girder_length_definition[1]

        # Member Shear Panel Sheeting Name
        clientObject.sheeting_name = sheeting_name

        # Member Shear Panel Fastening Arrangement
        clientObject.fastening_arrangement = fastening_arrangement.name

        # Member Shear Panel Panel Length
        clientObject.panel_length = panel_length

        # Member Shear Panel Beam Spacing
        clientObject.beam_spacing = beam_spacing

        # Member Shear Panel Coefficient K1
        clientObject.coefficient_k1 = coefficient_k1

        # Member Shear Panel Coefficient K2
        clientObject.coefficient_k2 = coefficient_k2

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Member Shear Panel to client model
        model.clientModel.service.set_member_shear_panel(clientObject)

    @staticmethod
    def Bracing(no: int = 1,
                name: str = '',
                girder_length_definition: list = [True],
                material_name: str = 'S235',
                diagonal_section: str = 'CHC 60.3x4.0',
                posts_section: str = 'CHC 76.1x4.0',
                modulus_of_elasticity: float = None,
                panel_length: float = 2.0,
                beam_spacing: float = 1.0,
                posts_spacing: float = 2.0,
                number_of_bracings: int = 2,
                diagonals_section_area: float = None,
                posts_section_area: float = None,
                comment: str = '',
                params: dict = None,
                model = Model):

        """
        Args:
            no (int): Member Shear Panel Tag
            name (str): User Defined Member Shear Panel Name
                if name == '':
                    user_defined_name_enabled = False (Automatic Name Assignment)
                else:
                    user_defined_name_enabled = True
                    name = User Defined Name
            girder_length_definition (list): Girder Length Definition List
                girder_length_definition[0] (bool): Activate/Deactivate Automatically Girder Length
                girder_length_definition[1] (float): Girder Length Value
            material_name (str): Material Name
            diagonal_section (str): Diagonal Section
            posts_section (str): Posts Section
            modulus_of_elasticity (float, optional): Modulus of Elasticity
            panel_length (float): Panel Length
            beam_spacing (float): Beam Spacing
            posts_spacing (float): Posts Spacing
            number_of_bracings (int): Number of Bracings
            diagonals_section_area (float, optional): Diagonals Section Area
            posts_section_area (float, optional): Posts Section Area
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client Model | Member Shear Panel
        clientObject = model.clientModel.factory.create('ns0:member_shear_panel')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Member Shear Panel No.
        clientObject.no = no

        # Member Shear Panel Definition Type
        clientObject.definition_type = MemberShearPanelDefinitionType.DEFINITION_TYPE_BRACING.name

        # Member Shear Panel User Defined Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Member Shear Panel Girder Length Definition
        if girder_length_definition[0]:
            clientObject.define_girder_length_automatically = True
        else:
            clientObject.define_girder_length_automatically = False
            clientObject.girder_length = girder_length_definition[1]

        # Member Shear Panel Material Name
        clientObject.material_name = material_name

        # Member Shear Panel Diagonal Section
        clientObject.diagonals_section_name = diagonal_section

        # Member Shear Panel Posts Section
        clientObject.posts_section_name = posts_section

        # Member Shear Panel Modulus of Elasticity
        clientObject.modulus_of_elasticity = modulus_of_elasticity

        # Member Shear Panel Panel Length
        clientObject.panel_length = panel_length

        # Member Shear Panel Beam Spacing
        clientObject.beam_spacing = beam_spacing

        # Member Shear Panel Posts Spacing
        clientObject.post_spacing = posts_spacing

        # Member Shear Panel Number of Bracing
        clientObject.number_of_bracings = number_of_bracings

        # Member Shear Panel Diagonals Section Area
        clientObject.diagonals_section_area = diagonals_section_area

        # Member Shear Panel Posts Section Area
        clientObject.posts_section_area = posts_section_area

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Member Shear Panel to client model
        model.clientModel.service.set_member_shear_panel(clientObject)

    @staticmethod
    def DefineSProv(no: int = 1,
                    name: str = '',
                    girder_length_definition: list = [True],
                    shear_panel_stiffness: float = 1000.0,
                    comment: str = '',
                    params: dict = None,
                    model = Model):

        """
        Args:
            no (int): Member Shear Panel Tag
            name (str): User Defined Member Shear Panel Name
                if name == '':
                    user_defined_name_enabled = False (Automatic Name Assignment)
                else:
                    user_defined_name_enabled = True
                    name = User Defined Name
            girder_length_definition (list): Girder Length Definition List
                girder_length_definition[0] (bool): Activate/Deactivate Automatically Girder Length
                girder_length_definition[1] (float): Girder Length Value
            shear_panel_stiffness (float): Shear Panel Stiffness
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client Model | Member Shear Panel
        clientObject = model.clientModel.factory.create('ns0:member_shear_panel')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Member Shear Panel No.
        clientObject.no = no

        # Member Shear Panel Definition Type
        clientObject.definition_type = MemberShearPanelDefinitionType.DEFINITION_TYPE_DEFINE_S_PROV.name

        # Member Shear Panel User Defined Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Member Shear Panel Stiffness
        clientObject.stiffness = shear_panel_stiffness

        # Member Shear Panel Girder Length Definition
        if girder_length_definition[0]:
            clientObject.define_girder_length_automatically = True
        else:
            clientObject.define_girder_length_automatically = False
            clientObject.girder_length = girder_length_definition[1]

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Member Shear Panel to client model
        model.clientModel.service.set_member_shear_panel(clientObject)

    @staticmethod
    def TrapeziodalSheetingAndBracing(
                    no: int = 1,
                    name: str = '',
                    sheeting_name: str = 'FI (+) 35/207 - 0.63 (b: 1) | DIN 18807 | Fischer Profil',
                    material_name: str = 'S235',
                    diagonals_section: str = 'CHC 60.3x3.2',
                    posts_section: str = 'CHC 76.1x4.0',
                    fastening_arrangement = MemberShearPanelFasteningArrangement.FASTENING_ARRANGEMENT_EVERY_RIB,
                    modulus_of_elasticity: float = None,
                    panel_length: float = 1.0,
                    girder_length_definition: list = [True],
                    beam_spacing: float = 2.0,
                    coefficient_k1: float = None,
                    coefficient_k2: float = None,
                    post_spacing: float = 3.0,
                    number_of_bracing: int = 2,
                    diagonals_section_area: float = None,
                    posts_section_area: float = None,
                    comment: str = '',
                    params: dict = None,
                    model = Model):

        """
        Args:
            no (int): Member Shear Panel Tag
            name (str): User Defined Member Shear Panel Name
                if name == '':
                    user_defined_name_enabled = False (Automatic Name Assignment)
                else:
                    user_defined_name_enabled = True
                    name = User Defined Name
            sheeting_name (str): Sheeting Name According to Library
            material_name (str): Material Name
            diagonals_section (str): Diagonals Section
            posts_section (str): Posts Section
            fastening_arrangement (enum): Fastening Arrangement Enumeration
            modulus_of_elasticity (float, optional): Modulus of Elasticity
            panel_length (float): Panel Length
            girder_length_definition (list): Girder Length Definition List
                girder_length_definition[0] (bool): Activate/Deactivate Automatically Girder Length
                girder_length_definition[1] (float): Girder Length Value
            beam_spacing (float): Beam Spacing
            coefficient_k1 (float, optional): Coefficient K1
            coefficient_k2 (float, optional): Coefficient K2
            post_spacing (float): Posts Spacing
            number_of_bracing (int): Number of Bracings
            diagonals_section_area (float, optional): Diagonals Section Area
            posts_section_area (float, optional): Posts Section Area
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client Model | Member Shear Panel
        clientObject = model.clientModel.factory.create('ns0:member_shear_panel')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Member Shear Panel No.
        clientObject.no = no

        # Member Shear Panel Definition Type
        clientObject.definition_type = MemberShearPanelDefinitionType.DEFINITION_TYPE_TRAPEZOIDAL_SHEETING_AND_BRACING.name

        # Member Shear Panel User Defined Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Member Shear Panel Sheeting Name
        clientObject.sheeting_name = sheeting_name

        # Member Shear Panel Material Name
        clientObject.material_name = material_name

        # Member Shear Panel Diagonals Section
        clientObject.diagonals_section_name = diagonals_section

        # Member Shear Panel Posts Section
        clientObject.posts_section_name = posts_section

        # Member Shear Panel Fastening Arrangement
        clientObject.fastening_arrangement = fastening_arrangement.name

        # Member Shear Panel Modulus of Elasticity
        clientObject.modulus_of_elasticity= modulus_of_elasticity

        # Member Shear Panel Panel Length
        clientObject.panel_length = panel_length

        # Member Shear Panel Beam Spacing
        clientObject.beam_spacing = beam_spacing

        # Member Shear Panel Girder Length Definition
        if girder_length_definition[0]:
            clientObject.define_girder_length_automatically = True
        else:
            clientObject.define_girder_length_automatically = False
            clientObject.girder_length = girder_length_definition[1]

        # Member Shear Panel Coefficients
        clientObject.coefficient_k1 = coefficient_k1
        clientObject.coefficient_k2 = coefficient_k2

        # Member Shear Panel Posts Spacings
        clientObject.post_spacing = post_spacing

        # Member Shear Panel Number of Bracing
        clientObject.number_of_bracings = number_of_bracing

        # Member Shear Panel Diagonals Section Area
        clientObject.diagonals_section_area = diagonals_section_area

        # Member Shear Panel Posts Section Area
        clientObject.posts_section_area = posts_section_area

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Member Shear Panel to client model
        model.clientModel.service.set_member_shear_panel(clientObject)



from RSTAB.enums import NodeType
from RSTAB.enums import NodeCoordinateSystemType
from RSTAB.enums import NodeReferenceType, ObjectTypes
from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes, ConvertStrToListOfInt
from math import pi

class Node():
    def __init__(self,
                 no: int = 1,
                 coordinate_X: float = 0.0,
                 coordinate_Y: float = 0.0,
                 coordinate_Z: float = 0.0,
                 comment: str = '',
                 params: dict = None,
                 model = Model):

        '''
         Args:
            no (int): Node Tag
            coordinate_X (float): X-Coordinate
            coordinate_Y (float): Y-Coordinate
            coordinate_Z (float): Z-Coordinate
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''
        # Client model | Node
        clientObject = model.clientModel.factory.create('ns0:node')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Node No.
        clientObject.no = no

        # Coordinates
        clientObject.coordinate_1 = coordinate_X
        clientObject.coordinate_2 = coordinate_Y
        clientObject.coordinate_3 = coordinate_Z

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)


        # Add Node to client model
        model.clientModel.service.set_node(clientObject)

    @staticmethod
    def Standard(
                 no: int = 1,
                 coordinate_system: list = None,
                 coordinate_system_type = NodeCoordinateSystemType.COORDINATE_SYSTEM_CARTESIAN,
                 comment: str = '',
                 params: dict = None,
                 model = Model):

        '''
         Args:
            no (int): Node Tag
            coordinate_system (list): Coordinate System
              for coordinate_system_type == NodeCoordinateSystemType.COORDINATE_SYSTEM_CARTESIAN;
                coordinate_system = [X, Y, Z]
              for coordinate_system_type == NodeCoordinateSystemType.COORDINATE_SYSTEM_X_CYLINDRICAL;
                coordinate_system = [X, R, θ]
              for coordinate_system_type == NodeCoordinateSystemType.COORDINATE_SYSTEM_Y_CYLINDRICAL;
                coordinate_system = [R, Ύ, θ]
              for coordinate_system_type == NodeCoordinateSystemType.COORDINATE_SYSTEM_Z_CYLINDRICAL;
                coordinate_system = [R, θ, Z]
              for coordinate_system_type == NodeCoordinateSystemType.COORDINATE_SYSTEM_POLAR:
                coordinate_system = [R, θ, φ]
            coordinate_system_type (enum): Coordinate System Type Enumeration
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''

        # Client model | Node
        clientObject = model.clientModel.factory.create('ns0:node')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Node No.
        clientObject.no = no

        # Node Type
        clientObject.type = NodeType.TYPE_STANDARD.name

        # Coordinates

        clientObject.coordinate_system_type = coordinate_system_type.name

        if len(coordinate_system) != 3:
            raise ValueError('WARNING: The coordinate system needs to be of length 3.')

        if not all(isinstance(x, (int, float)) for x in coordinate_system):
            raise Exception ('WARNING: Coordinate system should be type "int" or "float".')


        if clientObject.coordinate_system_type == "COORDINATE_SYSTEM_CARTESIAN":
            clientObject.coordinate_1 = coordinate_system[0]
            clientObject.coordinate_2 = coordinate_system[1]
            clientObject.coordinate_3 = coordinate_system[2]

        elif clientObject.coordinate_system_type == "COORDINATE_SYSTEM_X_CYLINDRICAL":
            clientObject.coordinate_1 = coordinate_system[0]
            clientObject.coordinate_2 = coordinate_system[1]
            clientObject.coordinate_3 = coordinate_system[2] * (pi/180)

        elif clientObject.coordinate_system_type == "COORDINATE_SYSTEM_Y_CYLINDRICAL":
            clientObject.coordinate_1 = coordinate_system[0]
            clientObject.coordinate_2 = coordinate_system[1]
            clientObject.coordinate_3 = coordinate_system[2] * (pi/180)

        elif clientObject.coordinate_system_type == "COORDINATE_SYSTEM_Z_CYLINDRICAL":
            clientObject.coordinate_1 = coordinate_system[0]
            clientObject.coordinate_2 = coordinate_system[1] * (pi/180)
            clientObject.coordinate_3 = coordinate_system[2]

        elif clientObject.coordinate_system_type == "COORDINATE_SYSTEM_POLAR":
            clientObject.coordinate_1 = coordinate_system[0]
            clientObject.coordinate_2 = coordinate_system[1] * (pi/180)
            clientObject.coordinate_3 = coordinate_system[2] * (pi/180)

        # Comment
        clientObject.comment = comment

        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Node to client model
        model.clientModel.service.set_node(clientObject)


    @staticmethod
    def BetweenTwoNodes(
                 no: int = 1,
                 start_node_no: int = 1,
                 end_node_no: int = 2,
                 node_reference = NodeReferenceType.REFERENCE_TYPE_L,
                 length_between_i_and_j: int = 1,
                 parameters = [True, 50],
                 offset_y: int = 0,
                 offset_z: int = 0,
                 comment: str = '',
                 params: dict = None,
                 model = Model):

        '''
        Args:
            no (int): Node Tag
            start_node_no (int): Start Node
            end_node_no (int): End Node
            node_reference (enum): Node Reference Enumeration
            length_between_i_and_j (float): Length between 2 Nodes
            parameters (list):
              if distance_from_start_relative:
                parameters = [True, %]
              if distance_from_start_absolute:
                parameters = [False, magnitude]
            offset_y (float): Offset in Y-Direction
            offset_z (float): Offset in Z-Direction
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''

        # Client model | Node
        clientObject = model.clientModel.factory.create('ns0:node')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Node No.
        clientObject.no = no

        # Node Type
        clientObject.type = NodeType.TYPE_BETWEEN_TWO_NODES.name

        # Start Node No.
        clientObject.between_two_nodes_start_node = start_node_no

        # End Node No.
        clientObject.between_two_nodes_end_node = end_node_no

         # Length between i and j
        clientObject.reference_type = node_reference.name

        clientObject.reference_object_projected_length = length_between_i_and_j

        # Distance between node k and start point
        if parameters[0]:
            clientObject.distance_from_start_relative = parameters[1]
        else:
            clientObject.distance_from_start_absolute = parameters[1]

        # Offset_local_y
        clientObject.offset_in_local_direction_y = offset_y

        # Offset_local_z
        clientObject.offset_in_local_direction_z = offset_z

        # Comment
        clientObject.comment = comment

        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Node to client model
        model.clientModel.service.set_node(clientObject)

    @staticmethod
    def BetweenTwoPoints(
                 no: int = 1,
                 start_point_x: float = 0.0,
                 start_point_y: float = 0.0,
                 start_point_z: float = 0.0,
                 end_point_x: float = 1.0,
                 end_point_y: float = 1.0,
                 end_point_z: float = 1.0,
                 node_reference = NodeReferenceType.REFERENCE_TYPE_L,
                 parameters = [True, 0.5],
                 offset_y: float = 0.0,
                 offset_z: float = 0.0,
                 comment: str = '',
                 params: dict = None,
                 model = Model):

        '''
        Args:
            no (int): Node Tag
            start_point_x (float): Start Point in X-Coordinate
            start_point_y (float): Start Point in Y-Coordinate
            start_point_z (float): Start Point in Z-Coordinate
            end_point_x (float): End Point in X-Coordinate
            end_point_y (float): End Point in Y-Coordinate
            end_point_z (float): End Point in Z-Coordinate
            node_reference (enum): Node Reference Enumeration
            parameters (list):
              if distance_from_start_relative:
                parameters = [True, %]
              if distance_from_start_absolute:
                parameters = [False, magnitude]
            offset_y (float): Offset in Y-Direction
            offset_z (float): Offset in Z-Direction
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''

        # Client model | Node
        clientObject = model.clientModel.factory.create('ns0:node')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Node No.
        clientObject.no = no

        # Node Type
        clientObject.type = NodeType.TYPE_BETWEEN_TWO_POINTS.name

        # Coordinates start point
        clientObject.between_two_points_start_point_coordinate_1= start_point_x
        clientObject.between_two_points_start_point_coordinate_2= start_point_y
        clientObject.between_two_points_start_point_coordinate_3= start_point_z

        # Coordinates end point

        clientObject.between_two_points_end_point_coordinate_1= end_point_x
        clientObject.between_two_points_end_point_coordinate_2= end_point_y
        clientObject.between_two_points_end_point_coordinate_3= end_point_z

        # Length between i and j

        clientObject.reference_type = node_reference.name

        # Distance between node k and start point

        if parameters[0]:
            clientObject.distance_from_start_relative = parameters[1]
        else:
            clientObject.distance_from_start_absolute = parameters[1]

        # offset local coordinates
        clientObject.offset_in_local_direction_y= offset_y
        clientObject.offset_in_local_direction_z= offset_z

        # Comment
        clientObject.comment = comment

        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Node to client model
        model.clientModel.service.set_node(clientObject)

    @staticmethod
    def OnMember(
                 no: int = 1,
                 member_number: int = 1,
                 node_reference = NodeReferenceType.REFERENCE_TYPE_L,
                 length_between_i_and_j: int = 1,
                 parameters = [True, 0.5],
                 comment: str = '',
                 params: dict = None,
                 model = Model):

        '''
         Args:
            no (int): Node Tag
            member_number (int): Member Tag
            node_reference (enum): Node Reference Enumeration
            length_between_i_and_j (int): Length between 2 Nodes
            parameters (list):
              if distance_from_start_relative:
                parameters = [True, %]
              if distance_from_start_absolute:
                parameters = [False, magnitude]
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''

        # Client model | Node
        clientObject = model.clientModel.factory.create('ns0:node')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Node No.
        clientObject.no = no

        # Member Reference No.
        clientObject.on_member_reference_member = member_number

        # Node Type
        clientObject.type = NodeType.TYPE_ON_MEMBER.name

        # Length between i and j

        clientObject.reference_type = node_reference.name

        clientObject.reference_object_projected_length = length_between_i_and_j

        # Distance between node k and start point

        if parameters[0]:
            clientObject.distance_from_start_relative = parameters[1]
        else:
            clientObject.distance_from_start_absolute = parameters[1]

        # Comment
        clientObject.comment = comment

        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Node to client model
        model.clientModel.service.set_node(clientObject)

    @staticmethod
    def DeleteNode(nodes_no: str = '1 2', model = Model):

        '''
        Args:
            nodes_no (str): Numbers of Nodes to be deleted
            model (RSTAB Class, optional): Model to be edited
        '''

        # Delete from client model
        for node in ConvertStrToListOfInt(nodes_no):
            model.clientModel.service.delete_object(ObjectTypes.E_OBJECT_TYPE_NODE.name, node)

    @staticmethod
    def GetNode(object_index: int = 1, model = Model):

        '''
        Args:
            obejct_index (int): Node Index
            model (RSTAB Class, optional): Model to be edited
        '''

        # Get Node from client model
        return model.clientModel.service.get_node(object_index)



from RSTAB.enums import MemberSupportNonlinearity
from RSTAB.initModel import Model, clearAttributes, ConvertToDlString, deleteEmptyAttributes
from RSTAB.dataTypes import inf

class MemberSupport():

    def __init__(self,
                 no: int = 1,
                 members: str = '',
                 spring_translation_x: float = 0.0,
                 spring_translation_y: float = 0.0,
                 spring_translation_z: list = [inf, MemberSupportNonlinearity.NONLINEARITY_NONE],
                 spring_rotation: float = 0.0,
                 spring_shear_x: float = 0.0,
                 spring_shear_y: float = 0.0,
                 spring_shear_z: float = 0.0,
                 member_rotational_restraint: int = 1,
                 comment: str = '',
                 params: dict = None,
                 model = Model):

        """
        Args:
            no (int): Member Support Tag
            members (str): Assigned Members
            spring_translation_x (float): Translational X Spring Constant
            spring_translation_y (float): Translational Y Spring Constant
            spring_translation_z (list): [Translational Z Spring Constant, Nonlinearity Type]
            spring_rotation (float): Rotational Spring Constant
            spring_shear_x (float): Shear X Spring Constant
            spring_shear_y (float): Shear Y Spring Constant
            spring_shear_z (float): Shear Z Spring Constant
            member_rotational_restraint (int): Member Rotational Restraint Number
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client model | Member Support
        clientObject = model.clientModel.factory.create('ns0:member_support')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Member Support No.
        clientObject.no = no

        # Assigned Members
        clientObject.members = ConvertToDlString(members)

        # Spring Translation
        clientObject.spring_rotation = spring_rotation
        clientObject.spring_translation_x = spring_translation_x
        clientObject.spring_translation_y = spring_translation_y
        clientObject.spring_translation_z = spring_translation_z[0]
        clientObject.nonlinearity_translational_z = spring_translation_z[1].name

        # Spring Shear
        clientObject.spring_shear_x = spring_shear_x
        clientObject.spring_shear_y = spring_shear_y
        clientObject.spring_shear_z = spring_shear_z

        # Member Rotational Restraint
        clientObject.member_rotational_restraint = member_rotational_restraint

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Member Support to client model
        model.clientModel.service.set_member_support(clientObject)



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes, ConvertToDlString
from RSTAB.enums import TimberEffectiveLengthsSupportType, TimberEffectiveLengthsEccentricityType, TimberEffectiveLengthsSupportTypeInY, \
    TimberEffectiveLengthsRestraintTypeAboutX, TimberEffectiveLengthsDeterminationType

class TimberEffectiveLengths():
    def __init__(self,
                 no: int = 1,
                 members: str = '1',
                 member_sets: str = '',
                 flexural_buckling_about_y: bool = True,
                 flexural_buckling_about_z: bool = True,
                 lateral_torsional_buckling: bool = True,
                 name: str = 'SEL1',
                 nodal_supports: list = [
                     [TimberEffectiveLengthsSupportType.SUPPORT_TYPE_FIXED_IN_Z_Y_AND_TORSION, True, TimberEffectiveLengthsEccentricityType.ECCENTRICITY_TYPE_NONE, \
                      TimberEffectiveLengthsSupportTypeInY.SUPPORT_STATUS_YES, TimberEffectiveLengthsRestraintTypeAboutX.SUPPORT_STATUS_YES, "1"],
                     [TimberEffectiveLengthsSupportType.SUPPORT_TYPE_FIXED_IN_Z_Y_AND_TORSION, True, TimberEffectiveLengthsEccentricityType.ECCENTRICITY_TYPE_NONE, \
                      TimberEffectiveLengthsSupportTypeInY.SUPPORT_STATUS_YES, TimberEffectiveLengthsRestraintTypeAboutX.SUPPORT_STATUS_YES, "2"],
                                        ],
                 factors: list = [
                     [1.0, 1.0, 1.0]
                                 ],
                 intermediate_nodes: bool = False,
                 different_properties: bool = True,
                 factors_definition_absolute: bool = False,
                 fire_design_different_buckling_factors: bool = False,
                 import_from_stability_analysis_enabled: bool = False,
                 determination_type = TimberEffectiveLengthsDeterminationType.DETERMINATION_EIGENVALUE_SOLVER,
                 comment: str = '',
                 params: dict = None,
                 model = Model):
        """
        Args:
            no (int): Timber Effective Length Tag
            members (str): Assigned Members
            member_sets (str): Assigned Member Sets
            flexural_buckling_about_y (bool): Flexural Buckling About Y Option
            flexural_buckling_about_z (bool): Flexural Buckling About Z Option
            lateral_torsional_buckling (bool): Lateral Torsional Buckling Option
            name (str): User Defined Effective Length Name
            nodal_supports (lst): Nodal Support Table Definition
                nodal_supports[i][0] (enum): Support Type Enumeration
                nodal_supports[i][1] (bool): Support in Z Option
                nodal_supports[i][2] (enum): Eccentricity Type Enumeration
                nodal_supports[i][3] (enum): Support Type in Y Enumeration
                nodal_supports[i][4] (enum): Restraint Type in X Enumeration
                nodal_supports[i][5] (str): Assigned Nodes
            factors (list of lists): Effective Length Factors
                factors[i][0] (float): Flexural Buckling in U Coefficient
                factors[i][1] (float): Flexural Buckling in V Coefficient
                factors[i][2] (float): Critical Moment
            intermediate_nodes (bool): Intermediate Nodes Option
            different_properties (bool): Different Properties Option
            factors_definition_absolute (bool): Absolute Factors Definition Option
            fire_design_different_buckling_factors (bool): Different Fire Design Buckling Factors
            import_from_stability_analysis_enabled (bool): Import From Stability Analysis Option
            determination_type (enum): Timber Effective Lengths Determination Type
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client Model | Types For Timber Design Effective Lengths
        clientObject = model.clientModel.factory.create('ns0:timber_effective_lengths')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Effective Lengths No.
        clientObject.no = no

        # Effective Lengths Assigned Members
        clientObject.members = ConvertToDlString(members)

        # Effective Lengths Assigned Member Sets
        clientObject.member_sets = ConvertToDlString(member_sets)

        # Effective Lengths Flexural Buckling About Y
        clientObject.flexural_buckling_about_y = flexural_buckling_about_y

        # Effective Lengths Flexural Buckling About Z
        clientObject.flexural_buckling_about_z = flexural_buckling_about_z

        # Effective Lengths Lateral Torsional Buckling
        clientObject.lateral_torsional_buckling = lateral_torsional_buckling

        # Effective Lengths Intermediate Points
        clientObject.intermediate_nodes = intermediate_nodes

        # Effective Lengths Different Properties
        clientObject.different_properties = different_properties

        # Effective Lengths Factors Definition Absolute
        clientObject.factors_definition_absolute = factors_definition_absolute

        # Effective Lengths User Defined Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Effective Lengths Nodal Supports
        clientObject.nodal_supports = model.clientModel.factory.create('ns0:array_of_timber_effective_lengths_nodal_supports')

        for i,j in enumerate(nodal_supports):
            mlvlp = model.clientModel.factory.create('ns0:timber_effective_lengths_nodal_supports_row')
            mlvlp.no = i+1
            mlvlp.row = model.clientModel.factory.create('ns0:timber_effective_lengths_nodal_supports')
            clearAttributes(mlvlp.row)
            mlvlp.row.support_type = nodal_supports[i][0].name
            mlvlp.row.support_in_z = nodal_supports[i][1]
            mlvlp.row.eccentricity_type = nodal_supports[i][2].name
            mlvlp.row.support_in_y_type = nodal_supports[i][3].name
            mlvlp.row.restraint_about_x_type = nodal_supports[i][4].name
            mlvlp.row.nodes = ConvertToDlString(nodal_supports[i][5])

            clientObject.nodal_supports.timber_effective_lengths_nodal_supports.append(mlvlp)

        # Effective Lengths Factors
        clientObject.factors = model.clientModel.factory.create('ns0:array_of_timber_effective_lengths_factors')

        for i,j in enumerate(factors):
            mlvlp_f = model.clientModel.factory.create('ns0:timber_effective_lengths_factors_row')
            mlvlp_f.no = i+1
            mlvlp_f.row = model.clientModel.factory.create('ns0:timber_effective_lengths_factors')
            clearAttributes(mlvlp_f.row)
            mlvlp_f.row.flexural_buckling_u = factors[i][0]
            mlvlp_f.row.flexural_buckling_v = factors[i][1]
            mlvlp_f.row.lateral_torsional_buckling = factors[i][2]

            clientObject.factors.timber_effective_lengths_factors.append(mlvlp_f)

        # Effective Lengths Fire Design
        clientObject.fire_design_different_buckling_factors = fire_design_different_buckling_factors

        # Effective Lengths Import From Stability Analysis Enabled
        clientObject.import_from_stability_analysis_enabled = import_from_stability_analysis_enabled

        # Effective Lengths Determination Type
        clientObject.determination_type = determination_type.name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Timber Effective Lengths to client model
        model.clientModel.service.set_timber_effective_lengths(clientObject)



import os
import sys
from re import findall, match
from shutil import copy

PROJECT_ROOT = os.path.abspath(os.path.join(
                  os.path.dirname(__file__),
                  os.pardir)
)
sys.path.append(PROJECT_ROOT)

from RSTAB.ImportExport.exports import ExportResultTablesToCSV

columns = 0

def __HTMLheadAndHeader(modelName, fileNames):
    '''
    Create header of the file.
    '''
    output = ['<head>',
              '<script src="htmlScript.js"></script>',
              '<meta http-equiv="Content-Type" content="text/html; charset=utf-8">',
              '<title>Results</title>',
              '<link rel="icon" sizes="32x32" href="favicon32.png">',
              '</head>',
              '<header>',
              '<link rel="stylesheet" href="htmlStyles.css">',
              '<div class="head">',
              '<img alt="Dlubal logo" width="77" height="77" src="favicon32.png" align="right">',
              '<h2>Result report</h2>',
              f'<p>Model: {modelName}</p>',
              '</div>',
              '<div class="head2">',
              '<input type="text" list="filter" id="fl" onfocus="this.value=\'\'" onchange="showPanel()" placeholder="Find result table">',
              '<datalist id="filter">']
    for f in fileNames:
        output.append(f'<option>{f}</option>')
    output += ['</datalist>',
               '<button class="button" onclick="switchButtonDown()">&#9664;</button>',
               '<button class="button" onclick="switchButtonUp()">&#9654;</button>',
               '</div>',
               '</header>',
               '',
               '<div class="tabContainer">']
    for f in fileNames:
        output.append(f'<iframe src="{f}.html" loading="lazy" frameBorder="0"></iframe>')
    return output

def __HTMLfooter():
    '''
    Create footer.
    '''
    return ['<script>atTheEnd()</script>',
            '</div>']

def __isEmpty(dividedLine):
    '''
    Returns True if all strings are empty.
    '''
    return not any (dividedLine)

def __numberOfPOsitiveItems(dividedLine):
    '''
    Get number of list items which are not empty.
    '''
    counter = 0
    for i in dividedLine:
        if i:
            counter += 1

    return counter

def __isWords(word):
    '''
    Check if 'word' consists of general words.
    '''
    isWords = False
    dividedWords = word.split(' ')

    for oneWord in dividedWords:
        if oneWord and bool(match(r"[A-Z]", oneWord[0])) and oneWord.isalpha():
            isWords = True
        else:
            break

    return isWords

def __tableHeader(dividedLine_1, dividedLine_2):
    '''
    Define html of header lines, rowspan and colspan.
    Parameters are lists of strings.
    '''
    global columns
    columns = max(len(dividedLine_1), len(dividedLine_2))
    output = ['<div class="tabPanel">',
              '<link rel="stylesheet" href="htmlStyles.css">',
              '<table class="responsive-table">',
              '<thead>',
              '<tr>']
    if __isEmpty(dividedLine_1):
        __emptyLine()
    else:
        for i in range(columns):
            if dividedLine_1[i] and not dividedLine_2[i]:
                output.append(f'<th rowspan="2">{dividedLine_1[i]}</th>')
            elif not dividedLine_1[i] and not dividedLine_2[i]:
                output.append('<th></th>')
            elif dividedLine_1[i] and dividedLine_2[i]:
                colspan = 1
                for ii in range(i+1, columns):
                    if ii == columns-1 and not dividedLine_1[ii] and __isWords(dividedLine_2[ii]):
                        break
                    if not dividedLine_1[ii] and dividedLine_2[ii]:
                        colspan += 1
                    else:
                        break
                output.append(f'<th colspan="{colspan}">{dividedLine_1[i]}</th>')
                i += colspan-1
            elif i==columns-1 and not dividedLine_1[i] and dividedLine_2[i]:
                output.append('<th></th>')
    output += ['</tr>', '<tr>']

    for y in range(columns):
        if not dividedLine_1[y] and not dividedLine_2[y]:
            output.append('<th></th>')
        elif dividedLine_1[y] and not dividedLine_2[y]:
            pass
        else:
            output.append(f'<th>{dividedLine_2[y]}</th>')

    output += ['</tr>', '</thead>']
    return output

def __tableSubHeader(dividedLine):
    '''
    Sub header; one liner; in the body of table
    '''
    return f'<th align="left" colspan="{columns}">{dividedLine}</th>'

def __emptyLine():
    '''
    Define html of empty lines
    '''
    return f'<th colspan="{columns}"></th>'

def __otherLines(dividedLine):
    '''
    Define html of other lines
    '''
    colspan = 1
    output = []
    for c in range(columns):
        if colspan == 1:
            sCheckIfDigit = dividedLine[c].replace('.','',1)
            sCheckIfDigit = sCheckIfDigit.replace(',','',1)
            sCheckIfDigit = sCheckIfDigit.replace('-','',1)
            sCheckIfDigit = sCheckIfDigit.replace('+','',1)
            align = 'left'
            tag = 'td'
            if c == 0:
                align = 'center'
            elif sCheckIfDigit.isdigit():
                align = 'right'
            elif bool(match(r"^\D+\.$", dividedLine[c])) or "|" in dividedLine[c]:
                for cc in range(c+1, columns-1):
                    if not dividedLine[cc]:
                        colspan += 1
                    else:
                        break
            else:
                if '<sub>' in dividedLine[c] or '[' in dividedLine[c]:
                    tag = 'th'
                    align = 'center'
            output.append(f'<{tag} colspan="{colspan}"align="{align}">{dividedLine[c]}</{tag}>')
        else:
            colspan -= 1
    return output

def __writeToFile(TargetFilePath, output):
    '''
    Write into html file
    '''
    with open(TargetFilePath, "w", encoding="utf-8") as f:
        for line in output:
            while '_' in line:
                beginId = line.find('_')
                endBySpace = line[beginId:].find(' ')
                endByArrow = line.rfind('<')
                if endBySpace == -1:
                    line = line[:beginId]+'<sub>'+line[beginId+1:endByArrow]+'</sub>'+line[endByArrow:]
                else:
                    endId = min(endBySpace + beginId, endByArrow)
                    line = line[:beginId]+'<sub>'+line[beginId+1:endId]+'</sub>'+line[endId:]
            while '^' in line:
                beginId = line.find('^')
                endBySpace = line[beginId:].find(' ')
                endByArrow = line.rfind('<')
                if endBySpace == -1:
                    line = line[:beginId]+'<sup>'+line[beginId+1:endByArrow]+'</sup>'+line[endByArrow:]
                else:
                    endId = min(endBySpace + beginId, endByArrow)
                    line = line[:beginId]+'<sup>'+line[beginId+1:endId]+'</sup>'+line[endId:]
            f.write(line+'\n')

def ExportResultTablesToHtml(TargetFolderPath: str, OpenBrowser: bool = True):
    '''
    This feature allows the user to create an output HTML file with the results.
    The results are in the same form as result tables in RSTAB.
    The output file is written in HTML and consists of embedded tabular data.
    It will also include dropdown menu with all tables/files.
    Result files are language dependent, so parsing based on strings is impossible.

    Args:
        TargetFolderPath (str): Destination path to the directory
    '''
    # Create collection of source files
    ExportResultTablesToCSV(TargetFolderPath)

    modelName = next(os.walk(TargetFolderPath))[1][0]
    dirList = os.listdir(os.path.join(TargetFolderPath, modelName))

    fileNames = []
    dirList.sort()

    for fileName in dirList:
        cats = fileName[:-4].split('_')

        fileNameCapitalized = ''
        for c in cats:
            if findall('[0-9]+', c):
                fileNameCapitalized += c+' '
            else:
                fileNameCapitalized += c.capitalize()+' '
        fileNameCapitalized = fileNameCapitalized[:-1]
        fileNames.append(fileNameCapitalized)

        with open(os.path.join(TargetFolderPath, modelName, fileName), mode='r', encoding='utf-8-sig') as f:
            lines = f.readlines()
            # 1st header always consists of first 2 lines
            line_1 = lines[0].split(';')
            line_1[-1] = line_1[-1].rstrip('\n')
            line_2 = lines[1].split(';')
            line_2[-1] = line_2[-1].rstrip('\n')

            output = __tableHeader(line_1, line_2)
            output.append('<tbody>')

            for line in lines[2:]:
                output.append('<tr>')
                dividedLine = line.split(';')
                dividedLine[-1] = dividedLine[-1].rstrip('\n')

                if __isEmpty(dividedLine):
                    # if empty line
                    output.append(__emptyLine())
                elif __numberOfPOsitiveItems(dividedLine) == 1 and dividedLine[0]:
                    # add empty line, these values doesn't make sence
                    output.append(__emptyLine())
                elif __numberOfPOsitiveItems(dividedLine) == 1 and dividedLine[1]:
                    # if only one string in the list, it is sub header consisting of only 1 line
                    output.append(__tableSubHeader(dividedLine[1]))
                else:
                    output += __otherLines(dividedLine)

            output += ['</tr>', '</tbody>', '</table>', '</div>']

        __writeToFile(os.path.join(TargetFolderPath, fileNameCapitalized+'.html'), output)

    indexOutput = ['<div class="tabContainer">']
    indexOutput += __HTMLfooter()
    indexOutput = __HTMLheadAndHeader(modelName, fileNames) + indexOutput

    # Copy basic files
    dirname = os.path.join(os.getcwd(), os.path.dirname(__file__))
    copy(os.path.join(dirname, 'htmlStyles.css'), TargetFolderPath)
    copy(os.path.join(dirname, 'htmlScript.js'), TargetFolderPath)
    copy(os.path.join(dirname, 'favicon32.png'), TargetFolderPath)

    with open(os.path.join(TargetFolderPath,'index.html'), "w", encoding="utf-8") as f:
        for line in indexOutput:
            f.write(line+'\n')

    if OpenBrowser:
        # Open result html page
        os.system(f"start {os.path.join(TargetFolderPath, 'index.html')}")


from RSTAB.initModel import Model

def CreateSectionList(name: str = "Favorites", model = Model):

    if isinstance(name, str):
        model.clientModel.service.create_my_section_list(name)
    else:
        raise ValueError("WARNING:Name of the section favorite list should be a string. Please kindly check the inputs.")

def AddSectionToMySectionList(list_name: str = "Favorites",
                             section_name: str = "IPE 300",
                             model = Model):

    if isinstance(list_name, str) and isinstance(section_name, str):
        model.clientModel.service.add_section_to_my_section_list(list_name, section_name)
    else:
        raise ValueError("WARNING:Name of the section favorite list and the section should be a string. Please kindly check the inputs.")

def DeleteSectionFromSectionList(list_name: str = "Favorites",
                                  section_name: str = "IPE 300",
                                  model = Model):

    if isinstance(list_name, str) and isinstance(section_name, str):
        model.clientModel.service.delete_section_from_my_section_list(list_name, section_name)
    else:
        raise ValueError("WARNING:Name of the section favorite list and the section should be a string. Please kindly check the inputs.")

def GetMySectionLists(model = Model):

    return model.clientModel.service.get_my_section_lists()

def DeleteSectionList(name: str = "Favorites", model = Model):

    if isinstance(name, str):
        model.clientModel.service.delete_my_section_list(name)
    else:
        raise ValueError("WARNING:Name of the section favorite list should be a string. Please kindly check the inputs.")

def CreateSectionFromRsectionFile(no: int = 1,
                                  file_path: str = "/rsection_file_path/",
                                  model = Model):

    if isinstance(no, int) and isinstance(file_path, str):
        model.clientModel.service.create_section_from_rsection_file(no, file_path)
    else:
        raise ValueError("WARNING: Type of file_path argument should be string and the type of the no argument should be integer. Please kindly check the inputs.")

def CreateSectionByName(id: int = 1,
                        material_id: str = "1",
                        name: str = 'my_section',
                        model = Model):

    if isinstance(id, int) and isinstance(material_id, int) and isinstance(name, str):
        model.clientModel.service.create_section_by_name(id, material_id, name)
    else:
        raise ValueError("WARNING: Type of id and material_id argument should be integer and the type of the name argument should be string. Please kindly check the inputs.")



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes, ConvertToDlString
from RSTAB.enums import MemberImperfectionType, MemberImperfectionDefinitionType
from RSTAB.enums import ImperfectionDirection, ImperfectionDirection, MemberImperfectionActiveCriterion

class MemberImperfection():

    def __init__(self,
                 no: int = 1,
                 imperfection_case: int = 1,
                 members: str ='1',
                 imperfection_type = MemberImperfectionType.IMPERFECTION_TYPE_INITIAL_SWAY,
                 definition_type = MemberImperfectionDefinitionType.DEFINITION_TYPE_ABSOLUTE,
                 imperfection_direction = ImperfectionDirection.IMPERFECTION_DIRECTION_LOCAL_Z,
                 parameters: list = [0.035],
                 comment: str = '',
                 params: dict = None,
                 model = Model):
        """
        Member Imperfection
        An imperfection case must be created before.

        Args:
            no (int): Member Imperfection Tag
            imperfection_case (int): Imperfection Case Number
            members (str): Assigned to Members
            imperfection_type (enum): Member Imperfection Type Enumeration
            definition_type (enum): Member Imperfection Definition Type Enumeration
            imperfection_direction (enum): Imperfection Direction Enumeration
            parameters (list): Parameters depending on Imperfection and Definition type
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client model | Imperfection Case
        clientObject = model.clientModel.factory.create('ns0:member_imperfection')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Member Imperfection No.
        clientObject.no = no

        # Member Imperfection Type
        clientObject.imperfection_type = imperfection_type.name

        # Assigned to Members No.
        clientObject.members = ConvertToDlString(members)

        # Imperfection Case
        clientObject.imperfection_case = imperfection_case

        # Definition Type
        clientObject.definition_type = definition_type.name

        # Imperfection Direction
        clientObject.imperfection_direction = imperfection_direction.name

        # Parameters
        if imperfection_type == MemberImperfectionType.IMPERFECTION_TYPE_INITIAL_SWAY:
            if definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_RELATIVE:
                clientObject.basic_value_relative = parameters[0]
            elif definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_ABSOLUTE:
                clientObject.basic_value_absolute = parameters[0]
            elif definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_EN_1993_1_1 or definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_EN_1992_1:
                clientObject.basic_value_relative = parameters[0]
                clientObject.height = parameters[1]
                clientObject.column_in_row = parameters[2]
                clientObject.reduction_factor_h = parameters[3]
                clientObject.reduction_factor_m = parameters[4]
                clientObject.initial_sway = parameters[5]
                clientObject.initial_sway_inverted = parameters[6]
            elif definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_EN_1995_1_1:
                clientObject.basic_value_relative = parameters[0]
                clientObject.height = parameters[1]
                clientObject.reduction_factor_h = parameters[2]
                clientObject.initial_sway = parameters[3]
                clientObject.initial_sway_inverted = parameters[4]
            elif definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_ANSI_CURRENT:
                clientObject.basic_value_coefficient = parameters[0]
                clientObject.standard_factor_enumeration = parameters[1].name
                clientObject.standard_factor_number = parameters[2]
            elif definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_GB_50017_2017_GRAVITY_LOAD:
                clientObject.basic_value_coefficient = parameters[0]
                clientObject.standard_factor_enumeration = parameters[1].name
                clientObject.standard_factor_number = parameters[2]
                clientObject.case_object = parameters[3]
            elif definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_CSA_CURRENT:
                clientObject.basic_value_coefficient = parameters[0]
            elif definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_CSA_GRAVITY_LOAD:
                clientObject.basic_value_coefficient = parameters[0]
                clientObject.case_object = parameters[1]
            elif definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_GB_50017_2017_CURRENT:
                clientObject.basic_value_relative = parameters[0]
                clientObject.height = parameters[1]
                clientObject.number_of_floors = parameters[2]
                clientObject.delta = parameters[3]
            elif definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_GB_50017_2017_GRAVITY_LOAD:
                clientObject.basic_value_coefficient = parameters[0]
                clientObject.number_of_floors = parameters[1]
                clientObject.case_object = parameters[2]
            else:
                assert False, 'Wrong MemberImperfectionDefinitionType for IMPERFECTION_TYPE_INITIAL_SWAY'

        elif imperfection_type == MemberImperfectionType.IMPERFECTION_TYPE_INITIAL_BOW:
            if definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_RELATIVE:
                clientObject.basic_value_relative = parameters[0]
            elif definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_ABSOLUTE:
                clientObject.basic_value_absolute = parameters[0]
            elif definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_EN_1993_1_1:
                clientObject.section_design = parameters[0].name
            elif definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_EN_1995_1_1:
                clientObject.basic_value_relative = parameters[0]
            elif definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_EN_1999_1_1:
                clientObject.section_design = parameters[0].name
            elif definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_ANSI_CURRENT:
                clientObject.basic_value_relative = parameters[0]
            elif definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_ANSI_GRAVITY_LOAD:
                clientObject.basic_value_relative = parameters[0]
            elif definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_CSA_CURRENT:
                clientObject.basic_value_relative = parameters[0]
            elif definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_CSA_GRAVITY_LOAD:
                clientObject.basic_value_relative = parameters[0]
            elif definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_GB_50017_2017:
                #clientObject. = parameters[0] # TODO: possibly missing parameter Buckling Curve = a - d
                pass
            else:
                assert False, 'Wrong MemberImperfectionDefinitionType for IMPERFECTION_TYPE_INITIAL_BOW'

        elif imperfection_type == MemberImperfectionType.IMPERFECTION_TYPE_INITIAL_BOW_AND_CRITERION:
            if definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_RELATIVE:
                clientObject.basic_value_relative = parameters[0]
                clientObject.active_criterion = parameters[1].name
                if clientObject.active_criterion == MemberImperfectionActiveCriterion.ACTIVITY_CRITERION_DEFINE:
                    clientObject.active_bow = parameters[2]
            elif definition_type == MemberImperfectionDefinitionType.DEFINITION_TYPE_ABSOLUTE:
                clientObject.basic_value_absolute = parameters[0]
                clientObject.active_criterion = parameters[1].name
                if clientObject.active_criterion == MemberImperfectionActiveCriterion.ACTIVITY_CRITERION_DEFINE:
                    clientObject.active_bow = parameters[2]
            else:
                assert False, 'Wrong MemberImperfectionDefinitionType for IMPERFECTION_TYPE_INITIAL_BOW_AND_CRITERION'
        else:
            assert False, 'Wrong MemberImperfectionDefinitionType'

        # Reference to List of Members
        clientObject.reference_to_list_of_members

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Member Imperfection to client model
        model.clientModel.service.set_member_imperfection(imperfection_case, clientObject)



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes
from RSTAB.enums import ObjectTypes, FormulaParameter

class Formula():

    def __init__(self,
                 object_type = ObjectTypes.E_OBJECT_TYPE_NODE,
                 object_no: int = 1,
                 attribute: str = 'coordinate_1',
                 formula: str = '2*3',
                 model = Model):
        '''
        Args:
            object_type (enum): Type of object for which parameters are searched where formulas are allowed
            object_no (int): Object number
            attribute (str): Attribute identificator
            formula (str): Formula in text format
            model (RSTAB Class, optional): Model to be edited
        '''

        # Client model | Object Location
        lc = model.clientModel.factory.create('ns0:object_location')
        opl = model.clientModel.factory.create('ns0:object_parameter_location')

        lstOfAttributes = self.GetListOfParametersFormulaAllowedFor(object_type, object_no, model)

        # Check if attribute is valid
        if attribute not in lstOfAttributes:
            raise ValueError('WARNING: It is not allowed to insert a formula for attribute:"'+attribute+'.')

        # Clears object attributes | Sets all attributes to None
        clearAttributes(lc)
        clearAttributes(opl)

        lc.type = object_type.name
        lc.no = object_no
        lc.parent_no = 0

        opl.attribute = attribute

        # Delete None attributes for improved performance
        deleteEmptyAttributes(lc)
        deleteEmptyAttributes(opl)

        # Add Formula to client model
        model.clientModel.service.set_formula(lc, opl, formula)


    @staticmethod
    def Get(object_type = ObjectTypes.E_OBJECT_TYPE_NODE,
            object_no: int = 1,
            attribute: str = 'coordinate_1',
            formula_param = FormulaParameter.FORMULA,
            model = Model):
        '''
        Args:
            object_type (enum): Type of object for which parameters are searched where formulas are allowed
            object_no (int): Object number
            attribute (str): Attribute identificator
            formula_param (enum): Formula parameters. You can choose from ALL, FORMULA, IS_VALID, CALCULATED_VALUE.
            model (RSTAB Class, optional): Model to be edited
        Returns:
            Formula for given type, number and attribute according to chosen formula_param.
        '''

        # Client model | Object Location
        lc = model.clientModel.factory.create('ns0:object_location')
        opl = model.clientModel.factory.create('ns0:object_parameter_location')

        lstOfAttributes = Formula.GetListOfParametersFormulaAllowedFor(object_type, object_no, model)

        # Check if attribute is valid
        if attribute not in lstOfAttributes:
            raise ValueError('WARNING: For the attribute "'+attribute+'" the formula is not allowed.')

        # Clears object attributes | Sets all attributes to None
        clearAttributes(lc)
        clearAttributes(opl)

        lc.type = object_type.name
        lc.no = object_no
        lc.parent_no = 0

        opl.attribute = attribute
        formulaDict = model.clientModel.service.get_formula(lc, opl)
        resultFormula = None

        # Return specific parameter of formula
        if formula_param == FormulaParameter.ALL:
            resultFormula = formulaDict
        elif formula_param == FormulaParameter.FORMULA:
            resultFormula = formulaDict['formula']
        elif formula_param == FormulaParameter.IS_VALID:
            resultFormula = formulaDict['is_valid']
        elif formula_param == FormulaParameter.CALCULATED_VALUE:
            resultFormula = formulaDict['calculated_value']
        else:
            raise ValueError("WARNING: Chosen formula_param doesn't exist.")

        return resultFormula

    @staticmethod
    def GetListOfParametersFormulaAllowedFor(
                     object_type = ObjectTypes.E_OBJECT_TYPE_NODE,
                     object_no: int = 1,
                     model = Model):
        '''
        Args:
            object_type (enum): Type of object for which parameters are searched where formulas are allowed
            object_no (int): Object number
            model (RSTAB Class, optional): Model to be edited
        Returns:
            attr_lst (list): List of attriburtes for which formulas are allowed
        '''

        # Client model | Object Location
        clientObject = model.clientModel.factory.create('ns0:object_location')

        # Clears object attributes | Sets all attributes to None
        clearAttributes(clientObject)

        clientObject.type = object_type.name
        clientObject.no = object_no
        clientObject.parent_no = 0
        lst = model.clientModel.service.get_list_of_parameters_formula_allowed_for(clientObject)

        attr_lst = []
        for i in lst[0]:
            attr_lst.append(i.attribute)

        return attr_lst



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes
from RSTAB.enums import *

class MemberEccentricity():
    def __init__(self,
                 no: int = 1,
                 name: str = '',
                 eccentricity_type = MemberEccentricitySpecificationType.TYPE_RELATIVE,
                 eccentricity_parameters: list = [MemberEccentricityHorizontalSectionAlignment.ALIGN_MIDDLE, MemberEccentricityVerticalSectionAlignment.ALIGN_MIDDLE],
                 transverse_offset_type = MemberEccentricityTransverseOffsetType.TRANSVERSE_OFFSET_TYPE_NONE,
                 transverse_offset_parameters: list = None,
                 axial_offset_active: bool = False,
                 hinge_location_at_node: bool = False,
                 comment: str = '',
                 params: dict = None,
                 model = Model):
        """
        Args:
            no (int): Eccentricity Tag
            name (str): User Defined Name
            eccentricity_type (enum): Member Eccentricity Specification Type Enumeration
            eccentricity_parameters (list): Eccentricity Parameters List
                for eccentricity_type == MemberEccentricitySpecificationType.TYPE_RELATIVE:
                    eccentricity_parameters = [horizontal_section_alignment, vertical_section_alignment]
                for eccentricity_type == MemberEccentricitySpecificationType.TYPE_ABSOLUTE:
                    eccentricity_parameters = [coordinate_system, offset_x, offset_y, offset_z]
                for eccentricity_type == MemberEccentricitySpecificationType.TYPE_RELATIVE_AND_ABSOLUTE:
                    eccentricity_parameters = [coordinate_system, offset_x, offset_y, offset_z, horizontal_section_alignment, vertical_section_alignment]
            transverse_offset_type (enum): Member Eccentricity Transverse Offset Type Enumeration
            transverse_offset_parameters (list): Transverse Offset Parameters List
                for transverse_offset_type == MemberEccentricityTransverseOffsetType.TRANSVERSE_OFFSET_TYPE_NONE:
                    transverse_offset_parameters = None
                for transverse_offset_type == MemberEccentricityTransverseOffsetType.TRANSVERSE_OFFSET_TYPE_FROM_SURFACE_THICKNESS:
                    transverse_offset_parameters = [transverse_offset_reference_surface, transverse_offset_vertical_alignment]
                for transverse_offset_type == MemberEccentricityTransverseOffsetType.TRANSVERSE_OFFSET_TYPE_FROM_MEMBER_SECTION:
                    transverse_offset_parameters = [transverse_offset_reference_member, transverse_offset_member_reference_node, transverse_offset_horizontal_alignment, transverse_offset_vertical_alignment]
            axial_offset_active (bool): Enable/Disable Axial Offset
            hinge_location_at_node (bool): Enable/Disable Hinge Location at Node
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client model | Member Eccentricity
        clientObject = model.clientModel.factory.create('ns0:member_eccentricity')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Member Eccentricity No.
        clientObject.no = no

        # User Defined Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Eccentricity Specification Type
        clientObject.specification_type = eccentricity_type.name

        if eccentricity_type.name == "TYPE_RELATIVE":
            clientObject.horizontal_section_alignment = eccentricity_parameters[0].name
            clientObject.vertical_section_alignment = eccentricity_parameters[1].name

        elif eccentricity_type.name == "TYPE_ABSOLUTE":
            clientObject.coordinate_system = eccentricity_parameters[0]
            clientObject.offset_x = eccentricity_parameters[1]
            clientObject.offset_y = eccentricity_parameters[2]
            clientObject.offset_z = eccentricity_parameters[3]

        elif eccentricity_type.name == "TYPE_RELATIVE_AND_ABSOLUTE":
            clientObject.coordinate_system = eccentricity_parameters[0]
            clientObject.offset_x = eccentricity_parameters[1]
            clientObject.offset_y = eccentricity_parameters[2]
            clientObject.offset_z = eccentricity_parameters[3]
            clientObject.horizontal_section_alignment = eccentricity_parameters[4].name
            clientObject.vertical_section_alignment = eccentricity_parameters[5].name

        else:
            raise ValueError("WARNING: Invalid eccentricity type.")

        # Transverse Offset Reference Type
        clientObject.transverse_offset_reference_type = transverse_offset_type.name

        if transverse_offset_type.name == "TRANSVERSE_OFFSET_TYPE_NONE":
            pass

        elif transverse_offset_type.name == "TRANSVERSE_OFFSET_TYPE_FROM_SURFACE_THICKNESS":
            clientObject.transverse_offset_reference_surface = transverse_offset_parameters[0]
            clientObject.transverse_offset_vertical_alignment = transverse_offset_parameters[1].name

        elif transverse_offset_type.name == "TRANSVERSE_OFFSET_TYPE_FROM_MEMBER_SECTION":
            clientObject.transverse_offset_reference_member = transverse_offset_parameters[0]
            clientObject.transverse_offset_member_reference_node = transverse_offset_parameters[1]
            clientObject.transverse_offset_horizontal_alignment = transverse_offset_parameters[2].name
            clientObject.transverse_offset_vertical_alignment = transverse_offset_parameters[3].name
        else:
            raise ValueError("WARNING: Invalid transverse offset type.")

        # Axial Offset Option
        clientObject.axial_offset_active = axial_offset_active

        # Hinge Location at Node Option
        clientObject.hinge_location_at_node = hinge_location_at_node

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Member Eccentricity to client model
        model.clientModel.service.set_member_eccentricity(clientObject)



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes
from RSTAB.enums import NodalReleaseTypePartialActivityAlong, NodalReleaseTypePartialActivityAround, NodalReleaseTypeLocalAxisSystemObjectType, NodalReleaseTypeReleaseNonlinearity
from RSTAB.dataTypes import inf

class NodalReleaseType():

    def __init__(self,
                 no: int = 1,
                 coordinate_system: str = "Local",
                 translational_release_n: float = inf,
                 translational_release_vy: float = inf,
                 translational_release_vz: float = inf,
                 rotational_release_mt: float = inf,
                 rotational_release_my: float = 0.0,
                 rotational_release_mz: float = 0.0,
                 translational_release_n_nonlinearity: list = [NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_NONE],
                 translational_release_vy_nonlinearity: list = [NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_NONE],
                 translational_release_vz_nonlinearity: list = [NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_NONE],
                 rotational_release_mt_nonlinearity: list = [NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_NONE],
                 rotational_release_my_nonlinearity: list = [NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_NONE],
                 rotational_release_mz_nonlinearity: list = [NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_NONE],
                 local_axis_system = NodalReleaseTypeLocalAxisSystemObjectType.LOCAL_AXIS_SYSTEM_OBJECT_TYPE_MEMBER,
                 local_axis_system_reference_object: int = 1,
                 name: str = None,
                 comment: str = '',
                 params: dict = None,
                 model = Model):

        '''
         Args:
            no (int): Nodal Release Type Tag
            coordinate_system (str/int): Coordinate System Selection
            translational_release_n (float): Spring Constant for Translation Release along X Direction
            translational_release_vy (float): Spring Constant for Translation Release along Y Direction
            translational_release_vz (float): Spring Constant for Translation Release along Z Direction
            rotational_release_mt (float): Spring Constant for Rotational Release around X Direction
            rotational_release_my (float): Spring Constant for Rotational Release around Y Direction
            rotational_release_mz (float): Spring Constant for Rotational Release around Z Direction
            translational_release_n_nonlinearity (list/list of lists): Nonlinearity Parameter for Translation Release along X Direction
            translational_release_vy_nonlinearity (list/list of lists): Nonlinearity Parameter for Translation Release along Y Direction
            translational_release_vz_nonlinearity (list/list of lists): Nonlinearity Parameter for Translation Release along Z Direction
                for translational_release_n/vy/vz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_PARTIAL_ACTIVITY:
                    translational_release_n/vy/vz_nonlinearity = [nonlinearity type Partial_Activity, negative zone, positive zone]
                    for negative/positive zone[0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_COMPLETE:
                        negative/positive zone = [negative/positive zone type, slippage]
                    for negative/positive zone[0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FIXED:
                        negative/positive zone = [negative/positive zone type, displacement, slippage]  (Note: Displacement must be greater than slippage)
                    for negative/positive zone[0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_FORCE/PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_FORCE:
                        negative/positive zone = [negative/positive zone type, force, slippage]
                for translational_release_n/vy/vz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_DIAGRAM:
                    translational_release_n/vy/vz_nonlinearity = [nonlinearity type Diagram, [symmetric(bool), NodalReleaseTypeDiagram Enumeration(start), NodalReleaseTypeDiagram Enumeration(end)], [[displacement, force],...]]
                for translational_release_n/vy/vz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FRICTION_DIRECTION_1/NONLINEARITY_TYPE_FRICTION_DIRECTION_2/NONLINEARITY_TYPE_FRICTION_DIRECTION_1_2:
                    translational_release_n/vy/vz_nonlinearity = [nonlinearity type Friction Direction, [friction coefficient(float)]]
                for translational_release_n/vy/vz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FRICTION_DIRECTION_1_PLUS_2:
                    translational_release_n/vy/vz_nonlinearity = [nonlinearity type Friction Direction, [friction coefficient 1(float), friction coefficient 2(float)]]
            rotational_release_mt_nonlinearity (list/list of lists): Nonlinearity Parameter for Rotational Release around X Direction
            rotational_release_my_nonlinearity (list/list of lists): Nonlinearity Parameter for Rotational Release around Y Direction
            rotational_release_mz_nonlinearity (list/list of lists): Nonlinearity Parameter for Rotational Release around Z Direction
                for rotational_release_mt/my/mz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_PARTIAL_ACTIVITY:
                    rotational_release_mt/my/mz_nonlinearity = [nonlinearity type Partial_Activity, negative zone, positive zone]
                    for negative/positive zone[0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_COMPLETE:
                        negative/positive zone = [negative/positive zone type, slippage]
                    for negative/positive zone[0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_FIXED:
                        negative/positive zone = [negative/positive zone type, rotation, slippage]
                    for negative/positive zone[0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_MOMENT/PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_MOMENT:
                        negative/positive zone = [negative/positive zone type, moment, slippage]
                for rotational_release_mt/my/mz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_DIAGRAM:
                    rotational_release_mt/my/mz_nonlinearity = [nonlinearity type Diagram, [symmetric(bool), NodalReleaseTypeDiagram Enumeration(start), NodalReleaseTypeDiagram Enumeration(end)], [[rotation, moment],...]]
            local_axis_system (enum): Nodal Release Local Axis System Enumeration
            local_axis_system_reference_object (int): Nodal Release Local Axis System Reference Object Number
            name (str, optional): User Defined Nodal Release Type Name
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''

        # Client model | Nodal Release Type
        clientObject = model.clientModel.factory.create('ns0:nodal_release_type')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Nodal Release Type No.
        clientObject.no = no

        # Nodal Release Type
        clientObject.coordinate_system = coordinate_system

         # Translational Release/Spring [N/m] N
        clientObject.axial_release_n = translational_release_n

        # Translational Release/Spring [N/m] Vy
        clientObject.axial_release_vy = translational_release_vy

        # Translational Release/Spring [N/m] Vz
        clientObject.axial_release_vz = translational_release_vz

        # Rotational Release/Spring [Nm/rad] Mt
        clientObject.moment_release_mt = rotational_release_mt

        # Rotational Release/Spring [Nm/rad] My
        clientObject.moment_release_my = rotational_release_my

        # Rotational Release/Spring [Nm/rad] Mz
        clientObject.moment_release_mz = rotational_release_mz

        # Translational Release N Nonlinearity

        # Nonlinearity Types None, Fixed if Negative N, Fixed if Positive N
        if translational_release_n_nonlinearity[0]== NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_NONE \
        or translational_release_n_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FAILURE_IF_NEGATIVE \
        or translational_release_n_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FAILURE_IF_POSITIVE :
            clientObject.axial_release_n_nonlinearity = translational_release_n_nonlinearity[0].name

        # Partial Activity
        elif translational_release_n_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_PARTIAL_ACTIVITY:
            clientObject.axial_release_n_nonlinearity = translational_release_n_nonlinearity[0].name

            # Negative Zone
            clientObject.partial_activity_along_x_negative_type = translational_release_n_nonlinearity[1][0].name

            if translational_release_n_nonlinearity[1][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_COMPLETE:
                clientObject.partial_activity_along_x_negative_slippage = translational_release_n_nonlinearity[1][1]

            elif translational_release_n_nonlinearity[1][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FIXED:
                clientObject.partial_activity_along_x_negative_displacement = translational_release_n_nonlinearity[1][1]
                clientObject.partial_activity_along_x_negative_slippage = translational_release_n_nonlinearity[1][2]

            elif translational_release_n_nonlinearity[1][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_FORCE:
                clientObject.partial_activity_along_x_negative_force = translational_release_n_nonlinearity[1][1]
                clientObject.partial_activity_along_x_negative_slippage = translational_release_n_nonlinearity[1][2]

            elif translational_release_n_nonlinearity[1][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_FORCE:
                clientObject.partial_activity_along_x_negative_force = translational_release_n_nonlinearity[1][1]
                clientObject.partial_activity_along_x_negative_slippage = translational_release_n_nonlinearity[1][2]

            # Positive Zone
            clientObject.partial_activity_along_x_positive_type = translational_release_n_nonlinearity[2][0].name

            if translational_release_n_nonlinearity[2][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_COMPLETE:
                clientObject.partial_activity_along_x_positive_slippage = translational_release_n_nonlinearity[2][1]

            elif translational_release_n_nonlinearity[2][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FIXED:
                clientObject.partial_activity_along_x_positive_displacement = translational_release_n_nonlinearity[2][1]
                clientObject.partial_activity_along_x_positive_slippage = translational_release_n_nonlinearity[2][2]

            elif translational_release_n_nonlinearity[2][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_FORCE:
                clientObject.partial_activity_along_x_positive_force = translational_release_n_nonlinearity[2][1]
                clientObject.partial_activity_along_x_positive_slippage = translational_release_n_nonlinearity[2][2]

            elif translational_release_n_nonlinearity[2][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_FORCE:
                clientObject.partial_activity_along_x_positive_force = translational_release_n_nonlinearity[2][1]
                clientObject.partial_activity_along_x_positive_slippage = translational_release_n_nonlinearity[2][2]

        elif translational_release_n_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_DIAGRAM:
            clientObject.axial_release_n_nonlinearity = translational_release_n_nonlinearity[0].name
            clientObject.diagram_along_x_symmetric = translational_release_n_nonlinearity[1][0]
            clientObject.diagram_along_x_is_sorted = True

            if translational_release_n_nonlinearity[1][0]:
                clientObject.diagram_along_x_start = translational_release_n_nonlinearity[1][1].name
                clientObject.diagram_along_x_end = translational_release_n_nonlinearity[1][1].name

            else:
                clientObject.diagram_along_x_start = translational_release_n_nonlinearity[1][1].name
                clientObject.diagram_along_x_end = translational_release_n_nonlinearity[1][2].name

            clientObject.diagram_along_x_table = model.clientModel.factory.create('ns0:nodal_release_type.diagram_along_x_table')

            for i,j in enumerate(translational_release_n_nonlinearity[2]):
                mlvlp = model.clientModel.factory.create('ns0:nodal_release_type_diagram_along_x_table_row')
                mlvlp.no = i+1
                mlvlp.row.displacement = translational_release_n_nonlinearity[2][i][0]
                mlvlp.row.force = translational_release_n_nonlinearity[2][i][1]

                clientObject.diagram_along_x_table.nodal_release_type_diagram_along_x_table.append(mlvlp)

        elif translational_release_n_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FRICTION_DIRECTION_1 \
        or translational_release_n_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FRICTION_DIRECTION_2 \
        or translational_release_n_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FRICTION_DIRECTION_1_2:
            clientObject.axial_release_n_nonlinearity = translational_release_n_nonlinearity[0].name
            clientObject.friction_coefficient_x = translational_release_n_nonlinearity[1][0]

        elif translational_release_n_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FRICTION_DIRECTION_1_PLUS_2:
            clientObject.axial_release_n_nonlinearity = translational_release_n_nonlinearity[0].name
            clientObject.friction_coefficient_xy = translational_release_n_nonlinearity[1][0]
            clientObject.friction_coefficient_xz = translational_release_n_nonlinearity[1][1]

        else:
            raise TypeError('Incorrect Nodal Release Nonlinearity Type')

        # Translational Release Vy Nonlinearity
        if translational_release_vy_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_NONE \
        or translational_release_vy_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FAILURE_IF_NEGATIVE \
        or translational_release_vy_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FAILURE_IF_POSITIVE:
            clientObject.axial_release_vy_nonlinearity = translational_release_vy_nonlinearity[0].name

        # Partial Activity
        elif translational_release_vy_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_PARTIAL_ACTIVITY:
            clientObject.axial_release_vy_nonlinearity = translational_release_vy_nonlinearity[0].name

            # Negative Zone
            clientObject.partial_activity_along_y_negative_type = translational_release_vy_nonlinearity[1][0].name

            if translational_release_vy_nonlinearity[1][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_COMPLETE:
                clientObject.partial_activity_along_y_negative_slippage = translational_release_vy_nonlinearity[1][1]

            elif translational_release_vy_nonlinearity[1][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FIXED:
                clientObject.partial_activity_along_y_negative_displacement = translational_release_vy_nonlinearity[1][1]
                clientObject.partial_activity_along_y_negative_slippage = translational_release_vy_nonlinearity[1][2]

            elif translational_release_vy_nonlinearity[1][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_FORCE:
                clientObject.partial_activity_along_y_negative_force = translational_release_vy_nonlinearity[1][1]
                clientObject.partial_activity_along_y_negative_slippage = translational_release_vy_nonlinearity[1][2]

            elif translational_release_vy_nonlinearity[1][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_FORCE:
                clientObject.partial_activity_along_y_negative_force = translational_release_vy_nonlinearity[1][1]
                clientObject.partial_activity_along_y_negative_slippage = translational_release_vy_nonlinearity[1][2]

            # Positive Zone
            clientObject.partial_activity_along_y_positive_type = translational_release_vy_nonlinearity[2][0].name

            if translational_release_vy_nonlinearity[2][0]== NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_COMPLETE:
                clientObject.partial_activity_along_y_positive_slippage = translational_release_vy_nonlinearity[2][1]

            elif translational_release_vy_nonlinearity[2][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FIXED:
                clientObject.partial_activity_along_y_positive_displacement = translational_release_vy_nonlinearity[2][1]
                clientObject.partial_activity_along_y_positive_slippage = translational_release_vy_nonlinearity[2][2]

            elif translational_release_vy_nonlinearity[2][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_FORCE:
                clientObject.partial_activity_along_y_positive_force = translational_release_vy_nonlinearity[2][1]
                clientObject.partial_activity_along_y_positive_slippage = translational_release_vy_nonlinearity[2][2]

            elif translational_release_vy_nonlinearity[2][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_FORCE :
                clientObject.partial_activity_along_y_positive_force = translational_release_vy_nonlinearity[2][1]
                clientObject.partial_activity_along_y_positive_slippage = translational_release_vy_nonlinearity[2][2]

        elif translational_release_vy_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_DIAGRAM:
            clientObject.axial_release_vy_nonlinearity = translational_release_vy_nonlinearity[0].name

            clientObject.diagram_along_y_symmetric = translational_release_vy_nonlinearity[1][0]
            clientObject.diagram_along_y_is_sorted = True

            if translational_release_vy_nonlinearity[1][0]:
                clientObject.diagram_along_y_start = translational_release_vy_nonlinearity[1][1].name
                clientObject.diagram_along_y_end = translational_release_vy_nonlinearity[1][1].name

            else:
                clientObject.diagram_along_y_start = translational_release_vy_nonlinearity[1][1].name
                clientObject.diagram_along_y_end = translational_release_vy_nonlinearity[1][2].name

            clientObject.diagram_along_y_table = model.clientModel.factory.create('ns0:nodal_release_type.diagram_along_y_table')

            for i,j in enumerate(translational_release_vy_nonlinearity[2]):
                mlvlp = model.clientModel.factory.create('ns0:nodal_release_type_diagram_along_y_table_row')
                mlvlp.no = i+1
                mlvlp.row = model.clientModel.factory.create('ns0:nodal_release_type_diagram_along_y_table')
                clearAttributes(mlvlp.row)
                mlvlp.row.displacement = translational_release_vy_nonlinearity[2][i][0]
                mlvlp.row.force = translational_release_vy_nonlinearity[2][i][1]

                clientObject.diagram_along_y_table.nodal_release_type_diagram_along_y_table.append(mlvlp)

        elif translational_release_vy_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FRICTION_DIRECTION_1 \
        or translational_release_vy_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FRICTION_DIRECTION_2 \
        or translational_release_vy_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FRICTION_DIRECTION_1_2:
            clientObject.axial_release_vy_nonlinearity = translational_release_vy_nonlinearity[0].name
            clientObject.friction_coefficient_y = translational_release_vy_nonlinearity[1][0]

        elif translational_release_vy_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FRICTION_DIRECTION_1_PLUS_2:
            clientObject.axial_release_vy_nonlinearity = translational_release_vy_nonlinearity[0].name
            clientObject.friction_coefficient_yx = translational_release_vy_nonlinearity[1][0]
            clientObject.friction_coefficient_yz = translational_release_vy_nonlinearity[1][1]

        else:
            raise TypeError('Incorrect Nodal Release Nonlinearity Type')

        # Translational Release Vz Nonlinearity
        if translational_release_vz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_NONE \
        or translational_release_vz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FAILURE_IF_NEGATIVE \
        or translational_release_vz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FAILURE_IF_POSITIVE :
            clientObject.axial_release_vz_nonlinearity = translational_release_vz_nonlinearity[0].name

        # Partial Activity
        elif translational_release_vz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_PARTIAL_ACTIVITY:
            clientObject.axial_release_vz_nonlinearity = translational_release_vz_nonlinearity[0].name

            # Negative Zone
            clientObject.partial_activity_along_z_negative_type = translational_release_vz_nonlinearity[1][0].name

            if translational_release_vz_nonlinearity[1][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_COMPLETE:
                clientObject.partial_activity_along_z_negative_slippage = translational_release_vz_nonlinearity[1][1]

            elif translational_release_vz_nonlinearity[1][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FIXED:
                clientObject.partial_activity_along_z_negative_displacement = translational_release_vz_nonlinearity[1][1]
                clientObject.partial_activity_along_z_negative_slippage = translational_release_vz_nonlinearity[1][2]

            elif translational_release_vz_nonlinearity[1][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_FORCE:
                clientObject.partial_activity_along_z_negative_force = translational_release_vz_nonlinearity[1][1]
                clientObject.partial_activity_along_z_negative_slippage = translational_release_vz_nonlinearity[1][2]

            elif translational_release_vz_nonlinearity[1][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_FORCE:
                clientObject.partial_activity_along_z_negative_force = translational_release_vz_nonlinearity[1][1]
                clientObject.partial_activity_along_z_negative_slippage = translational_release_vz_nonlinearity[1][2]

            # Positive Zone
            clientObject.partial_activity_along_z_positive_type = translational_release_vz_nonlinearity[2][0].name

            if translational_release_vz_nonlinearity[2][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_COMPLETE:
                clientObject.partial_activity_along_z_positive_slippage = translational_release_vz_nonlinearity[2][1]

            elif translational_release_vz_nonlinearity[2][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FIXED:
                clientObject.partial_activity_along_z_positive_displacement = translational_release_vz_nonlinearity[2][1]
                clientObject.partial_activity_along_z_positive_slippage = translational_release_vz_nonlinearity[2][2]

            elif translational_release_vz_nonlinearity[2][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_FORCE:
                clientObject.partial_activity_along_z_positive_force = translational_release_vz_nonlinearity[2][1]
                clientObject.partial_activity_along_z_positive_slippage = translational_release_vz_nonlinearity[2][2]

            elif translational_release_vz_nonlinearity[2][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_FORCE:
                clientObject.partial_activity_along_z_positive_force = translational_release_vz_nonlinearity[2][1]
                clientObject.partial_activity_along_z_positive_slippage = translational_release_vz_nonlinearity[2][2]

        elif translational_release_vz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_DIAGRAM:
            clientObject.axial_release_vz_nonlinearity = translational_release_vz_nonlinearity[0].name

            clientObject.diagram_along_z_symmetric = translational_release_vz_nonlinearity[1][0]
            clientObject.diagram_along_z_is_sorted = True

            if translational_release_vz_nonlinearity[1][0]:
                clientObject.diagram_along_z_start = translational_release_vz_nonlinearity[1][1].name
                clientObject.diagram_along_z_end = translational_release_vz_nonlinearity[1][1].name

            else:
                clientObject.diagram_along_z_start = translational_release_vz_nonlinearity[1][1].name
                clientObject.diagram_along_z_end = translational_release_vz_nonlinearity[1][2].name

            clientObject.diagram_along_z_table = model.clientModel.factory.create('ns0:member_hinge.diagram_along_z_table')

            for i,j in enumerate(translational_release_vz_nonlinearity[2]):
                mlvlp = model.clientModel.factory.create('ns0:member_hinge_diagram_along_z_table_row')
                mlvlp.no = i+1
                mlvlp.row.displacement = translational_release_vz_nonlinearity[2][i][0]
                mlvlp.row.force = translational_release_vz_nonlinearity[2][i][1]

                clientObject.diagram_along_z_table.member_hinge_diagram_along_z_table.append(mlvlp)

        elif translational_release_vz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FRICTION_DIRECTION_1 \
        or translational_release_vz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FRICTION_DIRECTION_2 \
        or translational_release_vz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FRICTION_DIRECTION_1_2:
            clientObject.axial_release_vz_nonlinearity = translational_release_vz_nonlinearity[0].name
            clientObject.friction_coefficient_z = translational_release_vz_nonlinearity[1][0]

        elif translational_release_vz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FRICTION_DIRECTION_1_PLUS_2:
            clientObject.axial_release_vz_nonlinearity = translational_release_vz_nonlinearity[0].name
            clientObject.friction_coefficient_zx = translational_release_vz_nonlinearity[1][0]
            clientObject.friction_coefficient_zy = translational_release_vz_nonlinearity[1][1]

        else:
            raise TypeError('Incorrect Nodal Release Nonlinearity Type')

        # Rotational Release Mt Nonlinearity
        if rotational_release_mt_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_NONE \
        or rotational_release_mt_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FAILURE_IF_NEGATIVE \
        or rotational_release_mt_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FAILURE_IF_POSITIVE :
            clientObject.moment_release_mt_nonlinearity = rotational_release_mt_nonlinearity[0].name

        # Partial Activity
        elif rotational_release_mt_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_PARTIAL_ACTIVITY:
            clientObject.moment_release_mt_nonlinearity = rotational_release_mt_nonlinearity[0].name

            # Negative Zone
            clientObject.partial_activity_around_x_negative_type = rotational_release_mt_nonlinearity[1][0].name

            if rotational_release_mt_nonlinearity[1][0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_COMPLETE:
                clientObject.partial_activity_around_x_negative_slippage = rotational_release_mt_nonlinearity[1][1]

            elif rotational_release_mt_nonlinearity[1][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FIXED:
                clientObject.partial_activity_around_x_negative_displacement = rotational_release_mt_nonlinearity[1][1]
                clientObject.partial_activity_around_x_negative_slippage = rotational_release_mt_nonlinearity[1][2]

            elif rotational_release_mt_nonlinearity[1][0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_MOMENT:
                clientObject.partial_activity_around_x_negative_moment = rotational_release_mt_nonlinearity[1][1]
                clientObject.partial_activity_around_x_negative_slippage = rotational_release_mt_nonlinearity[1][2]

            elif rotational_release_mt_nonlinearity[1][0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_MOMENT:
                clientObject.partial_activity_around_x_negative_moment = rotational_release_mt_nonlinearity[1][1]
                clientObject.partial_activity_around_x_negative_slippage = rotational_release_mt_nonlinearity[1][2]

            # Positive Zone
            clientObject.partial_activity_around_x_positive_type = rotational_release_mt_nonlinearity[2][0].name

            if rotational_release_mt_nonlinearity[2][0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_COMPLETE:
                clientObject.partial_activity_around_x_positive_slippage = rotational_release_mt_nonlinearity[2][1]

            elif rotational_release_mt_nonlinearity[2][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FIXED:
                clientObject.partial_activity_around_x_positive_displacement = rotational_release_mt_nonlinearity[2][1]
                clientObject.partial_activity_around_x_positive_slippage = rotational_release_mt_nonlinearity[2][2]

            elif rotational_release_mt_nonlinearity[2][0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_MOMENT:
                clientObject.partial_activity_around_x_positive_moment = rotational_release_mt_nonlinearity[2][1]
                clientObject.partial_activity_around_x_positive_slippage = rotational_release_mt_nonlinearity[2][2]

            elif rotational_release_mt_nonlinearity[2][0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_MOMENT:
                clientObject.partial_activity_around_x_positive_moment = rotational_release_mt_nonlinearity[2][1]
                clientObject.partial_activity_around_x_positive_slippage = rotational_release_mt_nonlinearity[2][2]

        elif rotational_release_mt_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_DIAGRAM:
            clientObject.moment_release_mt_nonlinearity = rotational_release_mt_nonlinearity[0].name

            clientObject.diagram_around_x_symmetric = rotational_release_mt_nonlinearity[1][0]
            clientObject.diagram_around_x_is_sorted = True

            if rotational_release_mt_nonlinearity[1][0]:
                clientObject.diagram_around_x_start = rotational_release_mt_nonlinearity[1][1].name
                clientObject.diagram_around_x_end = rotational_release_mt_nonlinearity[1][1].name

            else:
                clientObject.diagram_around_x_start = rotational_release_mt_nonlinearity[1][1].name
                clientObject.diagram_around_x_end = rotational_release_mt_nonlinearity[1][2].name

            clientObject.diagram_around_x_table = model.clientModel.factory.create('ns0:nodal_release_type.diagram_around_x_table')

            for i,j in enumerate(rotational_release_mt_nonlinearity[2]):
                mlvlp = model.clientModel.factory.create('ns0:nodal_release_type_diagram_around_x_table_row')
                mlvlp.no = i+1
                mlvlp.row = model.clientModel.factory.create('ns0:nodal_release_type_diagram_around_x_table')
                clearAttributes(mlvlp.row)
                mlvlp.row.rotation = rotational_release_mt_nonlinearity[2][i][0]
                mlvlp.row.moment = rotational_release_mt_nonlinearity[2][i][1]

                clientObject.diagram_around_x_table.nodal_release_type_diagram_around_x_table.append(mlvlp)

        else:
            raise TypeError('Incorrect Nodal Release Nonlinearity Type')

        # Rotational Release My Nonlinearity
        if rotational_release_my_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_NONE\
        or rotational_release_my_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FAILURE_IF_NEGATIVE \
        or rotational_release_my_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FAILURE_IF_POSITIVE :
            clientObject.moment_release_my_nonlinearity = rotational_release_my_nonlinearity[0].name

        # Partial Activity
        elif rotational_release_my_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_PARTIAL_ACTIVITY:
            clientObject.moment_release_my_nonlinearity = rotational_release_my_nonlinearity[0].name

            # Negative Zone
            clientObject.partial_activity_around_y_negative_type = rotational_release_my_nonlinearity[1][0].name

            if rotational_release_my_nonlinearity[1][0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_COMPLETE:
                clientObject.partial_activity_around_y_negative_slippage = rotational_release_my_nonlinearity[1][1]

            elif rotational_release_my_nonlinearity[1][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FIXED:
                clientObject.partial_activity_around_y_negative_displacement = rotational_release_my_nonlinearity[1][1]
                clientObject.partial_activity_around_y_negative_slippage = rotational_release_my_nonlinearity[1][2]

            elif rotational_release_my_nonlinearity[1][0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_MOMENT:
                clientObject.partial_activity_around_y_negative_moment = rotational_release_my_nonlinearity[1][1]
                clientObject.partial_activity_around_y_negative_slippage = rotational_release_my_nonlinearity[1][2]

            elif rotational_release_my_nonlinearity[1][0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_MOMENT:
                clientObject.partial_activity_around_y_negative_moment = rotational_release_my_nonlinearity[1][1]
                clientObject.partial_activity_around_y_negative_slippage = rotational_release_my_nonlinearity[1][2]

            # Positive Zone
            clientObject.partial_activity_around_y_positive_type = rotational_release_my_nonlinearity[2][0].name

            if rotational_release_my_nonlinearity[2][0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_COMPLETE:
                clientObject.partial_activity_around_y_positive_slippage = rotational_release_my_nonlinearity[2][1]

            elif rotational_release_my_nonlinearity[2][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FIXED:
                clientObject.partial_activity_around_y_positive_displacement = rotational_release_my_nonlinearity[2][1]
                clientObject.partial_activity_around_y_positive_slippage = rotational_release_my_nonlinearity[2][2]

            elif rotational_release_my_nonlinearity[2][0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_MOMENT:
                clientObject.partial_activity_around_y_positive_moment = rotational_release_my_nonlinearity[2][1]
                clientObject.partial_activity_around_y_positive_slippage = rotational_release_my_nonlinearity[2][2]

            elif rotational_release_my_nonlinearity[2][0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_MOMENT:
                clientObject.partial_activity_around_y_positive_moment = rotational_release_my_nonlinearity[2][1]
                clientObject.partial_activity_around_y_positive_slippage = rotational_release_my_nonlinearity[2][2]

        elif rotational_release_my_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_DIAGRAM:
            clientObject.moment_release_my_nonlinearity = rotational_release_my_nonlinearity[0].name

            clientObject.diagram_around_y_symmetric = rotational_release_my_nonlinearity[1][0]
            clientObject.diagram_around_y_is_sorted = True

            if rotational_release_my_nonlinearity[1][0]:
                clientObject.diagram_around_y_start = rotational_release_my_nonlinearity[1][1].name
                clientObject.diagram_around_y_end = rotational_release_my_nonlinearity[1][1].name

            else:
                clientObject.diagram_around_y_start = rotational_release_my_nonlinearity[1][1].name
                clientObject.diagram_around_y_end = rotational_release_my_nonlinearity[1][2].name

            clientObject.diagram_around_y_table = model.clientModel.factory.create('ns0:nodal_release_type.diagram_around_y_table')

            for i,j in enumerate(rotational_release_my_nonlinearity[2]):
                mlvlp = model.clientModel.factory.create('ns0:nodal_release_type_diagram_around_y_table_row')
                mlvlp.no = i+1
                mlvlp.row = model.clientModel.factory.create('ns0:nodal_release_type_diagram_around_y_table')
                clearAttributes(mlvlp.row)
                mlvlp.row.rotation = rotational_release_my_nonlinearity[2][i][0]
                mlvlp.row.moment = rotational_release_my_nonlinearity[2][i][1]

                clientObject.diagram_around_y_table.nodal_release_type_diagram_around_y_table.append(mlvlp)

        else:
            raise TypeError('Incorrect Nodal Release Nonlinearity Type')

        # Rotational Release Mz Nonlinearity
        if rotational_release_mz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_NONE \
        or rotational_release_mz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FAILURE_IF_NEGATIVE \
        or rotational_release_mz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_FAILURE_IF_POSITIVE :
            clientObject.moment_release_mz_nonlinearity = rotational_release_mz_nonlinearity[0].name

        # Partial Activity
        elif rotational_release_mz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_PARTIAL_ACTIVITY:
            clientObject.moment_release_mz_nonlinearity = rotational_release_mz_nonlinearity[0].name

            # Negative Zone
            clientObject.partial_activity_around_z_negative_type = rotational_release_mz_nonlinearity[1][0].name

            if rotational_release_mz_nonlinearity[1][0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_COMPLETE:
                clientObject.partial_activity_around_z_negative_slippage = rotational_release_mz_nonlinearity[1][1]

            elif rotational_release_mz_nonlinearity[1][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FIXED:
                clientObject.partial_activity_around_z_negative_displacement = rotational_release_mz_nonlinearity[1][1]
                clientObject.partial_activity_around_z_negative_slippage = rotational_release_mz_nonlinearity[1][2]

            elif rotational_release_mz_nonlinearity[1][0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_MOMENT:
                clientObject.partial_activity_around_z_negative_moment = rotational_release_mz_nonlinearity[1][1]
                clientObject.partial_activity_around_z_negative_slippage = rotational_release_mz_nonlinearity[1][2]

            elif rotational_release_mz_nonlinearity[1][0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_MOMENT:
                clientObject.partial_activity_around_z_negative_moment = rotational_release_mz_nonlinearity[1][1]
                clientObject.partial_activity_around_z_negative_slippage = rotational_release_mz_nonlinearity[1][2]

            # Positive Zone
            clientObject.partial_activity_around_z_positive_type = rotational_release_mz_nonlinearity[2][0].name

            if rotational_release_mz_nonlinearity[2][0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_COMPLETE:
                clientObject.partial_activity_around_z_positive_slippage = rotational_release_mz_nonlinearity[2][1]

            elif rotational_release_mz_nonlinearity[2][0] == NodalReleaseTypePartialActivityAlong.PARTIAL_ACTIVITY_TYPE_FIXED:
                clientObject.partial_activity_around_z_positive_displacement = rotational_release_mz_nonlinearity[2][1]
                clientObject.partial_activity_around_z_positive_slippage = rotational_release_mz_nonlinearity[2][2]

            elif rotational_release_mz_nonlinearity[2][0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_MOMENT:
                clientObject.partial_activity_around_z_positive_moment = rotational_release_mz_nonlinearity[2][1]
                clientObject.partial_activity_around_z_positive_slippage = rotational_release_mz_nonlinearity[2][2]

            elif rotational_release_mz_nonlinearity[2][0] == NodalReleaseTypePartialActivityAround.PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_MOMENT:
                clientObject.partial_activity_around_z_positive_moment = rotational_release_mz_nonlinearity[2][1]
                clientObject.partial_activity_around_z_positive_slippage = rotational_release_mz_nonlinearity[2][2]

        elif rotational_release_mz_nonlinearity[0] == NodalReleaseTypeReleaseNonlinearity.NONLINEARITY_TYPE_DIAGRAM:
            clientObject.moment_release_mz_nonlinearity = rotational_release_mz_nonlinearity[0].name

            clientObject.diagram_around_z_symmetric = rotational_release_my_nonlinearity[1][0]
            clientObject.diagram_around_z_is_sorted = True

            if rotational_release_mz_nonlinearity[1][0]:
                clientObject.diagram_around_z_start = rotational_release_mz_nonlinearity[1][1].name
                clientObject.diagram_around_z_end = rotational_release_mz_nonlinearity[1][1].name

            else:
                clientObject.diagram_around_z_start = rotational_release_mz_nonlinearity[1][1].name
                clientObject.diagram_around_z_end = rotational_release_mz_nonlinearity[1][2].name

            clientObject.diagram_around_z_table = model.clientModel.factory.create('ns0:nodal_release_type.diagram_around_z_table')

            for i,j in enumerate(rotational_release_mz_nonlinearity[2]):
                mlvlp = model.clientModel.factory.create('ns0:nodal_release_type_diagram_around_z_table_row')
                mlvlp.no = i+1
                mlvlp.row = model.clientModel.factory.create('ns0:nodal_release_type_diagram_around_z_table')
                clearAttributes(mlvlp.row)
                mlvlp.row.rotation = rotational_release_mz_nonlinearity[2][i][0]
                mlvlp.row.moment = rotational_release_mz_nonlinearity[2][i][1]

                clientObject.diagram_around_z_table.nodal_release_type_diagram_around_z_table.append(mlvlp)

        else:
            raise TypeError('Incorrect Nodal Release Nonlinearity Type')

        # Nodal Release Local Axis System Object Type
        clientObject.local_axis_system_object_type = local_axis_system.name

        # Nodal Release Local Axis System Reference Object
        clientObject.local_axis_system_reference_object = local_axis_system_reference_object

        # Nodal Release Type User defined name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Nodal Release Type to client model
        model.clientModel.service.set_nodal_release_type(clientObject)



from RSTAB.initModel import Model, SetAddonStatus, clearAttributes, ConvertToDlString, deleteEmptyAttributes
from RSTAB.enums import MemberTransverseStiffenerType, MemberTransverseStiffenerPosition, MemberTransverseStiffenerOffsetType, MemberTransverseStiffenerDefinitionType, AddOn

class MemberTransverseStiffener():

    # Member Transverse Component
    component = {'no' : 1,
                 'cantilever_l_c' : 0,
                 'stiffener_type' : MemberTransverseStiffenerType.STIFFENER_COMPONENT_TYPE_FLAT,
                 'position' : 1,
                 'position_type' : MemberTransverseStiffenerPosition.STIFFENER_COMPONENT_POSITION_DOUBLE_SIDED,
                 'multiple' : False,
                 'multiple_number' : 0,
                 'definition_type' : MemberTransverseStiffenerDefinitionType.DIMENSION_TYPE_OFFSET,
                 'multiple_offset_definition_type' : MemberTransverseStiffenerOffsetType.OFFSET_DEFINITION_TYPE_ABSOLUTE,
                 'multiple_offset' : 0,
                 'material' : 1,
                 'consider_stiffener' : True,
                 'thickness' : 0.005,
                 'width' : 0.02,
                 'height' : 0,
                 'non_rigid' : False,
                 'rigid' : False,
                 'width_b_u' : 0,
                 'height_h_u' : 0,
                 'thickness_t_u' : 0,
                 'thickness_s_u' : 0,
                 'width_b' : 0,
                 'thickness_t' : 0,
                 'column_section' : 0,
                 'section' : 0,
                 'full_warping_restraint' : False,
                 'user_defined_restraint' : False,
                 'user_defined_restraint_value' : 0,
                 'note' : 'comment'
            }

    def __init__(self,
                 no: int = 1,
                 members: str = "",
                 member_sets: str = "",
                 components: list = [component],
                 comment: str = '',
                 params: dict = None,
                 model = Model):
        """
        Args:
            no (int): Member Transverse Stiffener Tag
            members (str): Assigned Members
            member_sets (str): Assigned Member Sets
            components (list of components): List of components dictionary
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Check if Steel Design Add-on is ON.
        SetAddonStatus(model.clientModel, AddOn.steel_design_active)

        # Client Model | Member Transverse Stiffeners
        clientObject= model.clientModel.factory.create('ns0:member_transverse_stiffener')

        # Clear Object Atributes | Set All Atributes to None
        clearAttributes(clientObject)

        # Member Transverse Stiffeners No.
        clientObject.no = no

        # Assigned Members
        clientObject.members = ConvertToDlString(members)

        # Assigned Member Sets
        clientObject.member_sets = ConvertToDlString(member_sets)

        # Member Transverse Components
        clientObject.components = model.clientModel.factory.create('ns0:member_transverse_stiffener.components')

        for i in components:
            mlvlp = model.clientModel.factory.create('ns0:member_transverse_stiffener_components_row')
            mlvlp.no = i['no']
            mlvlp.row = model.clientModel.factory.create('ns0:member_transverse_stiffener_components')
            clearAttributes(mlvlp.row)
            mlvlp.row.stiffener_type = i['stiffener_type'].name
            mlvlp.row.position = i['position']
            mlvlp.row.position_type = i['position_type'].name
            mlvlp.row.multiple = i['multiple']
            mlvlp.row.multiple_number = i['multiple_number']
            mlvlp.row.multiple_offset_definition_type = i['multiple_offset_definition_type'].name
            mlvlp.row.multiple_offset = i['multiple_offset']
            mlvlp.row.material = i['material']
            mlvlp.row.consider_stiffener = i['consider_stiffener']
            mlvlp.row.thickness = i['thickness']
            mlvlp.row.width = i['width']
            mlvlp.row.height = i['height']
            mlvlp.row.non_rigid = i['non_rigid']
            mlvlp.row.rigid = i['rigid']
            mlvlp.row.width_b_u = i['width_b_u']
            mlvlp.row.height_h_u = i['height_h_u']
            mlvlp.row.thickness_t_u = i['thickness_t_u']
            mlvlp.row.thickness_s_u = i['thickness_s_u']
            mlvlp.row.width_b = i['width_b']
            mlvlp.row.thickness_t = i['thickness_t']
            mlvlp.row.column_section = i['column_section']
            mlvlp.row.section = i['section']
            mlvlp.row.full_warping_restraint = i['full_warping_restraint']
            mlvlp.row.user_defined_restraint = i['user_defined_restraint']
            mlvlp.row.user_defined_restraint_value = i['user_defined_restraint_value']
            mlvlp.row.note = i['note']
            mlvlp.row.cantilever_l_c = i['cantilever_l_c']
            mlvlp.row.definition_type = i['definition_type'].name

            clientObject.components.member_transverse_stiffener_components.append(mlvlp)

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key]= params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Member Definable Stffness to client model
        model.clientModel.service.set_member_transverse_stiffener(clientObject)



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes
from RSTAB.enums import ResultCombinationType, OperatorType, ActionLoadType

class ResultCombination():

    def __init__(self,
                 no: int = 1,
                 design_situation: int = 1,
                 combination_type = ResultCombinationType.COMBINATION_TYPE_GENERAL,
                 combination_items: list = [
                    [1, OperatorType.OPERATOR_OR, 1.2, ActionLoadType.LOAD_TYPE_TRANSIENT],
                    [2, OperatorType.OPERATOR_NONE, 1.6, ActionLoadType.LOAD_TYPE_TRANSIENT]
                    ],
                 generate_subcombinations: bool = False,
                 srss_combination: list = None,
                 name: str = '',
                 comment: str = '',
                 params: dict = None,
                 model = Model):

        '''
        Args:
            no (int): Result Combination tag
            design_situation (int): Assign Design Situation
            combination_type (enum): Result Combination Type Enumeration
            combination_items (list of lists): Combination Items Table
                combination_item[0](int) = case_object_item
                combination_item[1](enum) = operator_type enumeration
                combination_item[2](float) = case_object_factor
                combination_item[3](enum) = case_object_load_type enumeration
                combination_item[4](bool) = left_parenthesis (if parenthesis active in LoadCaseandCombination)
                combination_item[5](bool) = right_parenthesis (if parenthesis active in LoadCaseandCombination)
                combination_item[6](float) = group_factor (if left_parenthesis is True then value else None)
                combination_item[7](enum) = group_load_type (if right_parenthesis is True then enumeration else None)
            srss_combination (list, optional): SRSS Combination. If None then False.
                srss_combination = [srss_use_equivalent_linear_combination(bool), srss_extreme_value_sign(enum), srss_according_load_case_or_combination(int)]
            name (str, optional): User Defined Result Combination Name
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''

        # Client model | Result Combination
        clientObject = model.clientModel.factory.create('ns0:result_combination')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Result Combination No.
        clientObject.no = no

        # Result Combination Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Assigned Design Situation
        clientObject.design_situation = design_situation

        # Result Combination Type
        clientObject.combination_type = combination_type.name

        # Combination Items
        if combination_items:
            clientObject.items = model.clientModel.factory.create('ns0:result_combination.items')
            for i,j in enumerate(combination_items):
                rci = model.clientModel.factory.create('ns0:result_combination_items_row')
                rci.no = i+1
                rci.row = model.clientModel.factory.create('ns0:result_combination_items')
                clearAttributes(rci.row)
                rci.row.case_object_item = combination_items[i][0]
                rci.row.operator_type = combination_items[i][1].name
                rci.row.case_object_factor = combination_items[i][2]
                rci.row.case_object_load_type = combination_items[i][3].name
                if len(combination_items[i]) > 4:
                    rci.row.left_parenthesis = combination_items[i][4]
                    rci.row.right_parenthesis = combination_items[i][5]
                    if combination_items[i][4]:
                        rci.row.group_factor = combination_items[i][6]
                    if combination_items[i][5]:
                        rci.row.group_load_type = combination_items[i][7].name

                clientObject.items.result_combination_items.append(rci)

        # SRSS Combinations
        if not srss_combination:
            clientObject.srss_combination = False
        else:
            clientObject.srss_combination = True
            clientObject.srss_use_equivalent_linear_combination = srss_combination[0]
            clientObject.srss_extreme_value_sign = srss_combination[1].name
            if srss_combination[1].name == "EXTREME_VALUE_SIGN_ACCORDING_TO_LC_CO":
                clientObject.srss_according_load_case_or_combination = srss_combination[2]

        # Subcombinations
        clientObject.generate_subcombinations = generate_subcombinations

        clientObject.to_solve = True

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Result Combination to client model
        model.clientModel.service.set_result_combination(clientObject)



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes, ConvertStrToListOfInt
from RSTAB.enums import ObjectTypes

def ActionCombinationItem(model = Model, **kwargs):
    '''
    Set Action Combination Item.

    Args:
        model (RSTAB Class): Model to be edited
        **kwargs - Pass a keyworded, variable-length argument list. Following are all possible keywords:

        action_item, operator_type, left_parenthesis, right_parenthesis, group_factor, action_factor,
        action_load_type, group_load_type, action, is_leading, gamma, psi, xi, k_fi, c_esl, k_def,
        psi_0, psi_1, psi_2, fi, gamma_0, alfa, k_f, phi, omega_0, gamma_l_1, k_creep, gamma_n, j_2
    Usage:
        aci1 = ActionCombinationItem(Model, action_item=1, operator_type=OperatorType.OPERATOR_AND.name, action_factor=1.0, action=1)

    '''

    # Action Combination Item
    clientObject = model.clientModel.factory.create('ns0:action_combination_items')
    clearAttributes(clientObject)

    for key, value in kwargs.items():
        if key in clientObject.__keylist__:
            clientObject[key] = value

    deleteEmptyAttributes(clientObject)

    return clientObject


class ActionCombination():
    def __init__(self,
                 no: int = 1,
                 design_situation: int = 1,
                 action_combination_items: list = None,
                 name: str = '',
                 active: bool = True,
                 comment: str = '',
                 params: dict = None,
                 model = Model):
        '''
        Action Combination

        Attribute: Design Situation
        WARNING: Only design situations with an assigned combination wizard where a user-defined action combination is set are valid.

        Args:
            no (int): Action Combination Tag
            design_situation (int): Design Situation
            action_combination_items (list): Action Combination Items list
            name (str, optional): User Defined Action Combination Name
            active (bool, optional): Define if Active
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''

        # Client model | Action
        clientObject = model.clientModel.factory.create('ns0:action_combination')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Action Combination No.
        clientObject.no = no

        # Design Situation
        clientObject.design_situation = design_situation

        # Action Combination Items
        if action_combination_items:
            items = model.clientModel.factory.create('ns0:array_of_action_combination_items')
            for i,j in enumerate(action_combination_items):
                item = model.clientModel.factory.create('ns0:action_combination_items_row')
                item.no = i+1
                item.row = j

                items.action_combination_items.append(item)

            clientObject.items = items
        # Active
        clientObject.active = active

        # Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Clearing unused attributes
        deleteEmptyAttributes(clientObject)

        # Add Action to client model
        model.clientModel.service.set_action_combination(clientObject)

    @staticmethod
    def DeleteActionCombination(action_combination_no: str = '1 2', model = Model):
        '''
        Delete Action Combination object(s)

        Args:
            action_combination_no (str): Numbers of Action Combinations to be deleted
            model (RSTAB Class, optional): Model to be edited
        '''

        # Delete Action Combinations from client model
        for ac in ConvertStrToListOfInt(action_combination_no):
            model.clientModel.service.delete_object(ObjectTypes.E_OBJECT_TYPE_ACTION_COMBINATION.name, ac)

    @staticmethod
    def GetActionCombination(object_index: int = 1, model = Model):
        '''
        Get action Combination object

        Args:
            obejct_index (int): Action Combination Index
            model (RSTAB Class, optional): Model to be edited
        '''

        # Get Action Combination from client  model
        return model.clientModel.service.get_action_combination(object_index)



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes, ConvertStrToListOfInt
from RSTAB.enums import ObjectTypes

class Section():
    def __init__(self,
                 no: int = 1,
                 name: str = 'IPE 300',
                 material_no: int = 1,
                 comment: str = '',
                 params: dict = None,
                 model = Model):

        '''
        Args:
            no (int): Section Tag
            name (str): Name of Desired Section (As Named in RSTAB Database)
            material_no (int): Tag of Material assigned to Section
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''

        # Client model | Section
        clientObject = model.clientModel.factory.create('ns0:section')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Section No.
        clientObject.no = no

        # Section nNme
        clientObject.name = name

        # Material No.
        clientObject.material = material_no

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Section to client model
        model.clientModel.service.set_section(clientObject)

    @staticmethod
    def DeleteSection(sections_no: str = '1 2', model = Model):

        '''
        Args:
            sections_no (str): Numbers of Sections to be deleted
            model (RSTAB Class, optional): Model to be edited
        '''

        # Delete from client model
        for section in ConvertStrToListOfInt(sections_no):
            model.clientModel.service.delete_object(ObjectTypes.E_OBJECT_TYPE_SECTION.name, section)

    @staticmethod
    def GetSection(object_index: int = 1, model = Model):

        '''
        Args:
            obejct_index (int): Section Index
            model (RSTAB Class, optional): Model to be edited
        '''

        # Get Section from client model
        return model.clientModel.service.get_section(object_index)



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes, ConvertToDlString


class ConcreteUltimateConfiguration():

    def __init__(self,
                no: int = 1,
                name: str = 'ULS',
                members: str = 'All',
                member_sets: str = 'All',
                comment: str = '',
                params: dict = None,
                model = Model):
        """
        Args:
            no (int): Configuration Tag
            name (str): User Defined Name
            members (str): Assigned Members
            member_sets (str): Assigned Member Sets
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client model | Concrete Durabilities
        clientObject = model.clientModel.factory.create('ns0:concrete_design_uls_configuration')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Concrete Durability No.
        clientObject.no = no

        # User Defined Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Assigned Members
        if members == 'All':
            clientObject.assigned_to_all_members = True

        else:
            clientObject.assigned_to_all_members = False
            clientObject.assigned_to_members = ConvertToDlString(members)

        # Assigned Member Sets
        if member_sets == 'All':
            clientObject.assigned_to_all_member_sets = True

        else:
            clientObject.assigned_to_all_member_sets = False
            clientObject.assigned_to_member_sets = ConvertToDlString(member_sets)

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Global Parameter to client model
        model.clientModel.service.set_concrete_design_uls_configuration(clientObject)



from RSTAB.initModel import Model

def GetCOGTableParameters(results):
    '''
    Returns dict with attributes: base, row, and error.
    '''
    params = {'base': [], 'row': [], 'error': None}

    if not results:
        return params

    if results[0][0]:
        for section in results[0]:
            params['base'] = list(set(params['base'] + section.__keylist__))
            if 'rows' in section.__keylist__ and 'row' in section['rows']:
                for row in section['rows']['row']:
                    params['row'] = list(set(params['row'] + row.__keylist__))
            else:
                params['error'] = "Result table doesn't have attribute 'rows' or 'row'."

    return params

def ConvertCOGInfoToListOfDct(cog):
    '''
    Args:
        cog (dict): Dictionary object of center of gravity and objects information
    Returns:
        List of dictionaries. Each dictionary corresponds to one line in center of gravity and objects information table.
    '''
    if not cog:
        return ''

    params = GetCOGTableParameters(cog)
    lstOfDct = []

    for c in cog[0]:
        dct1 = {}
        if 'title' in params['base']:
            try:
                dct1['title'] = float(c['title'])
            except:
                try:
                    dct1['title'] = c['title']
                except:
                    pass

        if 'rows' in params['base']:

            for row in c['rows'][0]:
                dct = {}
                dct.update(dct1)
                for y in params['row']:
                    try:
                        dct[y] = float(row[y])
                    except:
                        try:
                            dct[y] = row[y]
                        except:
                            pass
                lstOfDct.append(dct)

    if params['error']:
        return lstOfDct.append({'error': params['error']})

    return lstOfDct

class ObjectsInfo():

    @staticmethod
    def AllInfo(objects: list = None,
                 model = Model):
        '''
        Args:
            objects (list of lists): List of Selected Objects
                objects = [[ObjectTypes enumeration, Object number], ...]   (e.g. [[ObjectTypes.E_OBJECT_TYPE_MEMBER,1], [ObjectTypes.E_OBJECT_TYPE_MEMBER, 2]])
            model (class, optional): Model instance

        Returns:
            List of dictionaries.
        '''

        if objects:
            cog_objects = {'center_of_gravity_objects' : []}

            for obj in objects:
                cog_objects["center_of_gravity_objects"].append({
                    "type": obj[0].name,
                    "no": obj[1]
                })
            return ConvertCOGInfoToListOfDct(model.clientModel.service.get_center_of_gravity_and_objects_info(cog_objects))

        else:
            raise AssertionError('WARNING! Please give at lease one object.')

    @staticmethod
    def CenterofGravity(objects: list = None,
                        model = Model):
        '''
        Args:
            objects (list of lists): List of Selected Objects
                objects = [[ObjectTypes enumeration, Object number], ...]   (e.g. [[ObjectTypes.E_OBJECT_TYPE_MEMBER,1], [ObjectTypes.E_OBJECT_TYPE_MEMBER, 2]])
            model (class, optional): Model instance

        Returns:
            List of dictionary.
        '''

        allList = ObjectsInfo.AllInfo(objects, model)
        cog_Info = {}
        for item in allList:
            if item['title'] == 'Center of Gravity':
               cog_Info[item.get('name')] = item.get('value')

        return cog_Info

    @staticmethod
    def AllSelectedObjectsInfo(objects: list = None,
                           model = Model):
        '''
        Args:
            objects (list of lists): List of Selected Objects
                objects = [[ObjectTypes enumeration, Object number], ...]   (e.g. [[ObjectTypes.E_OBJECT_TYPE_MEMBER,1], [ObjectTypes.E_OBJECT_TYPE_MEMBER, 2]])
            model (class, optional): Model instance

        Returns:
            List of dictionary.
        '''

        allList = ObjectsInfo.AllInfo(objects, model)
        cog_Info = {}
        for item in allList:
            if item['title'] == 'Information About All Selected Objects':
               cog_Info[item.get('name')] = item.get('value')

        return cog_Info

    @staticmethod
    def MembersInfo(objects: list = None,
                model = Model):
        '''
        Args:
            objects (list of lists): List of Selected Objects
                objects = [[ObjectTypes enumeration, Object number], ...]   (e.g. [[ObjectTypes.E_OBJECT_TYPE_MEMBER,1], [ObjectTypes.E_OBJECT_TYPE_MEMBER, 2]])
            model (class, optional): Model instance

        Returns:
            List of dictionary.
        '''

        allList = ObjectsInfo.AllInfo(objects, model)
        cog_Info = {}
        for item in allList:
            if item['title'] == 'Information About Members':
               cog_Info[item.get('name')] = item.get('value')

        return cog_Info

    @staticmethod
    def SurfacesInfo(objects: list = None,
                 model = Model):
        '''
        Args:
            objects (list of lists): List of Selected Objects
                objects = [[ObjectTypes enumeration, Object number], ...]   (e.g. [[ObjectTypes.E_OBJECT_TYPE_MEMBER,1], [ObjectTypes.E_OBJECT_TYPE_MEMBER, 2]])
            model (class, optional): Model instance

        Returns:
            List of dictionary.
        '''

        allList = ObjectsInfo.AllInfo(objects, model)
        cog_Info = {}
        for item in allList:
            if item['title'] == 'Information About Surfaces':
               cog_Info[item.get('name')] = item.get('value')

        return cog_Info

    @staticmethod
    def SolidsInfo(objects: list = None,
               model = Model):
        '''
        Args:
            objects (list of lists): List of Selected Objects
                objects = [[ObjectTypes enumeration, Object number], ...]   (e.g. [[ObjectTypes.E_OBJECT_TYPE_MEMBER,1], [ObjectTypes.E_OBJECT_TYPE_MEMBER, 2]])
            model (class, optional): Model instance

        Returns:
            List of dictionary.
        '''

        allList = ObjectsInfo.AllInfo(objects, model)
        cog_Info = {}
        for item in allList:
            if item['title'] == 'Information About Solids':
               cog_Info[item.get('name')] = item.get('value')

        return cog_Info

    @staticmethod
    def EnvelopeSize(objects: list = None,
                     model = Model):
        '''
        Args:
            objects (list of lists): List of Selected Objects
                objects = [[ObjectTypes enumeration, Object number], ...]   (e.g. [[ObjectTypes.E_OBJECT_TYPE_MEMBER,1], [ObjectTypes.E_OBJECT_TYPE_MEMBER, 2]])
            model (class, optional): Model instance

        Returns:
            List of dictionary.
        '''

        allList = ObjectsInfo.AllInfo(objects, model)
        cog_Info = {}
        for item in allList:
            if item['title'] == 'Envelope Size':
               cog_Info[item.get('name')] = item.get('value')

        return cog_Info
